/*-----------------------------------------------------------------------------
 * $RCSfile: Envelope.c,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if (!defined(_WIN32)) || defined(_DEBUG)
static  const   char    CVS_ID[]        =       "$Id: Envelope.c,v 1.80 2002/10/10 09:47:19 bandou Exp $";
#endif  /* _DEBUG */

#include "Envelope.h"
#include "Block.h"
#include "XMLNamespace.h"

#include <OpenSOAP/Locale.h>

#include <string.h>
#include <ctype.h>

/* SOAP Envelope Element Name */
static
const char SOAPEnvelopeElementName[] = "Envelope";
/* */
static
const char SOAPHeaderElementName[] = "Header";
/* */
static
const char SOAPBodyElementName[] = "Body";

/*
 */
typedef struct {
	const char *soapVersion;
	const char *soapEnvURI;
	const char *soapEnvDefPrefix;
	const char *soapActorName;
	const char *soapActorNext;
} SoapVersionEnvValues;

typedef const SoapVersionEnvValues *SoapVersionEnvValuesMapConstIterator;

static
const
SoapVersionEnvValues SOAP_VERSION_ENVURI_MAP[] = {
	{ "1.1",
	  "http://schemas.xmlsoap.org/soap/envelope/",
	  "SOAP-ENV",
	  "actor",
	  "http://schemas.xmlsoap.org/soap/actor/next"},
	{ "1.2",
	  "http://www.w3.org/2001/06/soap-envelope",
	  "env",
	  "role",
	  "http://www.w3.org/2002/06/soap-envelope/role/next"},
	{ NULL, NULL, NULL, NULL, NULL}
};

/*
=begin
= OpenSOAP Envelope class
=end
 */

static
int
OpenSOAPEnvelopeReleaseMembers(OpenSOAPEnvelopePtr
                               /* [in, out] */ soap_env) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (soap_env) {
        OpenSOAPStringRelease(soap_env->version);

        ret = OpenSOAPXMLElmReleaseMembers((OpenSOAPXMLElmPtr)soap_env);
    }

    return ret;
}

static
int
OpenSOAPEnvelopeFree(OpenSOAPObjectPtr /* [in, out] */ obj) {
    int ret = OpenSOAPEnvelopeReleaseMembers((OpenSOAPEnvelopePtr)obj);

    if (OPENSOAP_SUCCEEDED(ret)) {
        free(obj);
    }

    return ret;
}

static
int
OpenSOAPEnvelopeInitialize(OpenSOAPEnvelopePtr /* [in, out] */ soap_env) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (soap_env) {
        soap_env->version = NULL;
        soap_env->header  = NULL;
        soap_env->body    = NULL;

        ret = OpenSOAPXMLElmInitialize((OpenSOAPXMLElmPtr)soap_env,
                                       OpenSOAPEnvelopeFree);
    }

    return ret;
}



/*
=begin
--- function#OpenSOAPEnvelopeCreate(soap_env)
    OpenSOAP Envelope instance create
    
    :Parameters
      :OpenSOAPEnvelopePtr * [out] ((|soap_env|))
        Strage Buffer of OpenSOAP Envelope Pointer
    :Return value
      :int
	    Return Code Is ErrorCode
    :Note
      SOAP Envelope ̈̊mۂƍ\̂̏s܂B
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeCreate(OpenSOAPEnvelopePtr *soap_env) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (soap_env) {
        ret = OPENSOAP_MEM_BADALLOC;
        *soap_env = malloc(sizeof(OpenSOAPEnvelope));
        if (*soap_env) {
            ret = OpenSOAPEnvelopeInitialize(*soap_env);

            if (OPENSOAP_FAILED(ret)) {
                free(*soap_env);
                *soap_env = NULL;
            }
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeSetValuesMB(soapEnv, soapVer, envURI, envPrefix)
    OpenSOAP Envelope set variables.

    :Parameters
      :[out] const cahr * ((||soapEnv|))
        OpenSOAP Envelope Pointer.
      :[in]  const char * ((|soapVer|))
	    SOAP Version
      :[in]  const char * ((|envURI|))
        SOAP Envelope namespace URI.
      :[in]  const char * ((|envPrefix|))
        SOAP Envelope namespace prefix.
    :Return value
      :int 
	    Error code.
=end
 */
static
int
OpenSOAPEnvelopeSetValuesMB(/* [out] */ OpenSOAPEnvelopePtr soapEnv,
							/* [in]  */ const char *soapVer,
							/* [in]  */ const char *envURI,
							/* [in]  */ const char *envPrefix) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;
	if (soapEnv && soapVer && envURI && envPrefix) {
		OpenSOAPXMLElmPtr  rootElm  = (OpenSOAPXMLElmPtr)(soapEnv);
		OpenSOAPXMLNodePtr rootNode = (OpenSOAPXMLNodePtr)(rootElm);
		ret = OpenSOAPStringCreateWithMB(SOAPEnvelopeElementName,
										 &(rootNode->name));
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = OpenSOAPStringCreateWithMB(soapVer,
											 &(soapEnv->version));
					
			if (OPENSOAP_SUCCEEDED(ret)) {
				ret = OpenSOAPXMLNamespaceCreateMB(envURI,
												   envPrefix,
												   &(rootNode
													 ->thisNamespace));
				if (OPENSOAP_SUCCEEDED(ret)) {
					rootNode->thisNamespace->definedElement
						= rootElm;
					rootElm->definedXMLNsList
						= rootNode->thisNamespace;
				}
			}
		}
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeCreateMB(soapVer, envPrefix, soapEnv)
    OpenSOAP Envelope instance create(MB)

    :Parameters
      :const cahr * [in] ((||soapVer|))
        SOAP Virsion
      :const char * [in] ((|envPrefix|))
        SOAP Envelope namespace prefix.
      :OpenSOAPEnvelopePtr * [out] ((|soapEnv|))
        Strage Buffer of OpenSOAP Envelope Pointer
    :Return value
      :int 
	    Return Code Is ErrorCode
    :Note
      (2001/08/28)
	  SOAP Envelope쐬̈̊mۂsASOAP Versionɂ
	  eX萔̑s܂B
      soapVer      NULL ȊȌꍇ "1.1" or "1.2"
                    NULL ̏ꍇ "1.1" ƓB
      envPrefix    NULL ȊȌꍇ͓eɏ]B
                    NULL ̏ꍇsoapVerɂȉɏ]B
                    "1.1" SOAP-ENV
                    "1.2" env
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeCreateMB(const char *soapVer,
                         const char *envPrefix,
                         OpenSOAPEnvelopePtr *soapEnv) {
#if 1
	int ret = OPENSOAP_PARAMETER_BADVALUE;
	if (soapEnv) {
		SoapVersionEnvValuesMapConstIterator i = SOAP_VERSION_ENVURI_MAP;
		if (soapVer) {
			for (; i->soapVersion && strcmp(soapVer, i->soapVersion) != 0;
				 ++i) {
			}
		}
		if (i->soapVersion) {
			ret = OpenSOAPEnvelopeCreate(soapEnv);
			if (OPENSOAP_SUCCEEDED(ret)) {
				if (!envPrefix) {
					envPrefix = i->soapEnvDefPrefix;
				}
				ret = OpenSOAPEnvelopeSetValuesMB(*soapEnv,
												  i->soapVersion,
												  i->soapEnvURI,
												  envPrefix);
			}
			if (OPENSOAP_FAILED(ret)) {
				OpenSOAPEnvelopeRelease(*soapEnv);
				*soapEnv = NULL;
			}
		}
	}

	return ret;
#else
    static const char *SOAP_VERSIONS[] = {
        "1.1",
        "1.2",
        NULL
    };
    int ret = OPENSOAP_PARAMETER_BADVALUE;
    const char **s_ver_i = SOAP_VERSIONS;

    if (!soapVer) {
        soapVer = SOAP_VERSIONS[0];
    }

    for (; *s_ver_i && strcmp(soapVer, *s_ver_i) != 0; ++s_ver_i) {
    }
    if (!*s_ver_i) {
        return ret;
    }

    ret = OpenSOAPEnvelopeCreate( soapEnv );
    if ( OPENSOAP_FAILED(ret) ){
        return ret;
    }

    ret = OpenSOAPStringCreate( &((*soapEnv)->envElm.super.name) );
    if( OPENSOAP_FAILED(ret) ){
        return ret;
    }

    ret = OpenSOAPStringSetStringMB((*soapEnv)->envElm.super.name,
									SOAPEnvelopeElementName);
    if(ret == OPENSOAP_PARAMETER_BADVALUE){
        return ret;
    }


    if( !(strcmp( soapVer, "1.1" )) ){
		if( envPrefix == NULL ) { 
			ret = OpenSOAPXMLNamespaceCreateMB( "http://schemas.xmlsoap.org/soap/envelope/",
												"SOAP-ENV",
												&((*soapEnv)->envElm.super.thisNamespace)); 
		}
		else {
			ret = OpenSOAPXMLNamespaceCreateMB( "http://schemas.xmlsoap.org/soap/envelope/",
												envPrefix,
												&((*soapEnv)->envElm.super.thisNamespace)); 
		}

        if(OPENSOAP_FAILED(ret)){
            return ret;
        }
    }
    else{
		if( envPrefix == NULL ) { 
			ret = OpenSOAPXMLNamespaceCreateMB( "http://www.w3.org/2001/06/soap-envelope",
												"env",
												&((*soapEnv)->envElm.super.thisNamespace));
		}
		else {
			ret = OpenSOAPXMLNamespaceCreateMB( "http://www.w3.org/2001/06/soap-envelope",
												envPrefix,
												&((*soapEnv)->envElm.super.thisNamespace));
		}
        if(OPENSOAP_FAILED(ret)){
            return ret;
        }
    }
	
    ret = OpenSOAPStringCreate( &((*soapEnv)->version) );
    if( OPENSOAP_FAILED(ret) ){
        return ret;
    }

    ret = OpenSOAPStringSetStringMB((*soapEnv)->version, soapVer);
    if( OPENSOAP_FAILED(ret)){
        return ret;
    }

    (*soapEnv)->envElm.super.thisNamespace->definedElement = (OpenSOAPXMLElmPtr)soapEnv;
    (*soapEnv)->envElm.definedXMLNsList = (*soapEnv)->envElm.super.thisNamespace;

    return ret;
#endif	
}

/*
=begin
--- function#OpenSOAPStringConvertWCtoMB(wcStr, memAllocator, len, mbStr)
    Convert const wchar_t * to char *

    :Parameters
      :[in]  const wchar_t * ((|wcStr|))
        wide character string.
      :[in]  char * ( * ((|memAllocator|)) ) (size_t)
        memory allocator
      :[out] size_t * ((|len|))
        size of string buffer. If NULL, no effect.
      :[out] char ** ((|mbStr|))
        string stored buffer pointer.
    :Return value
      :type int
        Error code.
=end
 */
static
int
OpenSOAPStringConvertWCtoMB(/* [in]  */ const wchar_t *wcStr,
							/* [in]  */  char *(*memAllocator)(size_t),
							/* [out] */ size_t *len,
							/* [out] */ char **mbStr) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;
	if (mbStr) {
		ret = OPENSOAP_NO_ERROR;
		*mbStr = NULL;
		if (wcStr) {
			OpenSOAPStringPtr convStr = NULL;
			ret = OpenSOAPStringCreateWithWC(wcStr, &convStr);
			if (OPENSOAP_SUCCEEDED(ret)) {
				/* */
				ret = OpenSOAPStringGetStringMBWithAllocator(convStr,
															 memAllocator,
															 len,
															 mbStr);
				/* release */
				OpenSOAPStringRelease(convStr);
			}
		}
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeCreateWC(soapVer, envPrefix, soapEnv)
    OpenSOAP Envelope instance create(WC)

    :Parameters
      :const cahr * [in] ((||soapVer|))
        SOAP Virsion
      :const char * [in] ((|envPrefix|))
        SOAP Envelope namespace prefix.
      :OpenSOAPEnvelopePtr * [out] ((|soapEnv|))
        Strage Buffer of OpenSOAP Envelope Pointer
    :Return value
      :type int
        Error code.
    :Note
      (2001/08/28)
	  SOAP Envelope쐬̈̊mۂsASOAP Versionɂ
	  eX萔̑s܂B
      soapVer      NULL ȊȌꍇ "1.1" or "1.2"
                    NULL ̏ꍇ "1.1" ƓB
      envPrefix    NULL ȊȌꍇ͓eɏ]B
                    NULL ̏ꍇsoapVerɂȉɏ]B
                    "1.1" SOAP-ENV
                    "1.2" env
      
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeCreateWC(const wchar_t *soapVer,
                         const wchar_t *envPrefix,
                         OpenSOAPEnvelopePtr *soapEnv) {
	char *soapVerMB = NULL;
	int ret = OpenSOAPStringConvertWCtoMB(soapVer, NULL, NULL, &soapVerMB);
	if (OPENSOAP_SUCCEEDED(ret)) {
		char *envPrefixMB = NULL;
		ret = OpenSOAPStringConvertWCtoMB(envPrefix, NULL, NULL, &envPrefixMB);
		if (OPENSOAP_SUCCEEDED(ret)) {
			/* */
			ret = OpenSOAPEnvelopeCreateMB(soapVerMB, envPrefixMB, soapEnv);
			/* */
			free(envPrefixMB);
		}
		/* */
		free(soapVerMB);
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeCreateString(soapVer, envPrefix, soapEnv)
    OpenSOAP Envelope instance create(OpenSOAPString)

    :Parameters
      :[in]  OpenSOAPStringPtr ((||soapVer|))
        SOAP Virsion
      :[in]  OpenSOAPStringPtr ((|envPrefix|))
        SOAP Envelope namespace prefix.
      :[out] OpenSOAPEnvelopePtr * ((|soapEnv|))
        Strage Buffer of OpenSOAP Envelope Pointer
    :Return value
      :type int
        Error code.
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeCreateString(/* [in]  */ OpenSOAPStringPtr soapVer,
							 /* [in]  */ OpenSOAPStringPtr envPrefix,
							 /* [out] */ OpenSOAPEnvelopePtr *soapEnv) {
	char *soapVerMB = NULL;
	int ret = soapVer ?
		OpenSOAPStringGetStringMBWithAllocator(soapVer,
											   NULL,
											   NULL,
											   &soapVerMB)
		: OPENSOAP_NO_ERROR;
	if (OPENSOAP_SUCCEEDED(ret)) {
		char *envPrefixMB = NULL;
		ret = envPrefix ?
			OpenSOAPStringGetStringMBWithAllocator(envPrefix,
												   NULL,
												   NULL,
												   &envPrefixMB)
			: OPENSOAP_NO_ERROR;
		if (OPENSOAP_SUCCEEDED(ret)) {
			/* */
			ret = OpenSOAPEnvelopeCreateMB(soapVerMB, envPrefixMB, soapEnv);
			/* */
			free(envPrefixMB);
		}
		/* */
		free(soapVerMB);
	}

	return ret;
}

/*
 */
static
const unsigned char *
searchXMLDocBody(/* [in] */ const unsigned char *xmldocBegin,
				 /* [in] */ size_t xmldocSize) {
	static const unsigned char XML_PROC_END[] = "?>";
	static const unsigned char *XML_PROC_END_END
		= XML_PROC_END + sizeof(XML_PROC_END) - 1;

	const unsigned char *ret = xmldocBegin;
	if (xmldocBegin && xmldocSize) {
		const unsigned char *xmldocEnd = xmldocBegin + xmldocSize;
		const unsigned char *xmlProcEndI = XML_PROC_END;
		for (; xmlProcEndI != XML_PROC_END_END
				 && ret != xmldocEnd; ++ret) {
			if (*ret == *xmlProcEndI) {
				++xmlProcEndI;
			}
			else {
				xmlProcEndI = XML_PROC_END;
			}
		}
		if (ret == xmldocEnd) {
			ret = xmldocBegin;
		}
		for (; ret != xmldocEnd && isspace(*ret); ++ret) {
		}
	}

	return ret;
}

/*
 */
static
int
OpenSOAPByteArrayRemoveXMLProcAttr(/* [in]  */ OpenSOAPByteArrayPtr xmldoc,
								   /* [out] */ char **buf,
								   /* [out] */ size_t *bufSz) {
	static const char DEFAULT_XML_PROC[] = "<?xml version=\"1.0\"?>";
	static const size_t DEFAULT_XML_PROC_SIZE
		= sizeof(DEFAULT_XML_PROC) - 1;
	static const char SGML_TAG_BEGIN = '<';
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (xmldoc && buf && bufSz) {
		const unsigned char *xmldocBegin = NULL;
		size_t xmldocSize = 0;
		*buf = NULL;
		*bufSz = 0;
		ret = OpenSOAPByteArrayGetBeginSizeConst(xmldoc,
												 &xmldocBegin,
												 &xmldocSize);
		if (OPENSOAP_SUCCEEDED(ret)) {
			const unsigned char *xmldocBody 
				= searchXMLDocBody(xmldocBegin, xmldocSize);
			size_t xmldocBodySize = xmldocSize - (xmldocBody - xmldocBegin);
			if (xmldocBodySize == 0
				|| *xmldocBody != SGML_TAG_BEGIN) {
				ret = OPENSOAP_XML_NOTXMLDOCUMENT;
			} 
			else {
				*bufSz = DEFAULT_XML_PROC_SIZE + xmldocBodySize;
				*buf = malloc(*bufSz);
				if (*buf) {
					memcpy(*buf, DEFAULT_XML_PROC, DEFAULT_XML_PROC_SIZE);
					memcpy(*buf + DEFAULT_XML_PROC_SIZE,
						   xmldocBody, xmldocBodySize);
				}
				else {
					*bufSz = 0;
					ret = OPENSOAP_MEM_BADALLOC;
				}
			}
		}
	}
	
	return ret;
}

/*
=begin
--- function#OOpenSOAPLibxmlDocCreateCharEncoding(chEnc, bAry, doc, rootNode)
    create char encoding
    
    :Parameters
      :[in]  const cahr * ((|chEnc|))
        character encoding
      :[in]  OpenSOAPByteArrayPtr ((|bAry|))
        OpenSOAP ByteArray
      :[out] xmlDocPtr * ((|doc|))
        libxml document buffer.
      :[out] xmlNodePtr * ((|rootNode|))
        libxml document root node buffer.
    :Return value
      :type int
        error code.
    :Note
      
=end
*/
static
int
OpenSOAPLibxmlDocCreateCharEncoding(const char *chEnc,
									OpenSOAPByteArrayPtr bAry,
									xmlDocPtr	*doc,
									xmlNodePtr	*rootNode) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (bAry && doc && rootNode) {
		OpenSOAPByteArrayPtr utf8BAry = NULL;
		*doc = NULL;
		*rootNode = NULL;
		/* change b_arry to "UTF-8" */
		ret = OpenSOAPByteArrayCreate(&utf8BAry);
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = OpenSOAPStringConvertCharEncoding(chEnc,
													bAry,
													"UTF-8",
													utf8BAry);
			if (OPENSOAP_SUCCEEDED(ret)) {
				char *utf8Buf = NULL;
				size_t utf8BufSize = 0;
				ret = OpenSOAPByteArrayRemoveXMLProcAttr(utf8BAry,
														 &utf8Buf,
														 &utf8BufSize);
				if (OPENSOAP_SUCCEEDED(ret)) {
					*doc = xmlParseMemory(utf8Buf, utf8BufSize);
					if (!*doc) {
						ret = OPENSOAP_XML_BADMAKEDOCUMENT;
					}
					else {
						/* check document (root node) */
						*rootNode = xmlDocGetRootElement(*doc);
						if (!*rootNode) {
							xmlFreeDoc(*doc);
							*doc = NULL;
							ret = OPENSOAP_XML_BADMAKEDOCUMENT;
						}
					}
					free(utf8Buf);
				}
			}
			OpenSOAPByteArrayRelease(utf8BAry);
		}
	}
	
	return ret;
}

/*
=begin
--- function#OOpenSOAPLibxmlDocCreateCharEncoding(chEnc, bAry, doc, rootNode)
    create char encoding
    
    :Parameters
      :[in]  const cahr * ((|chEnc|))
        character encoding
      :[in]  OpenSOAPByteArrayPtr ((|bAry|))
        OpenSOAP ByteArray
      :[out] xmlDocPtr * ((|doc|))
        libxml document buffer.
      :[out] xmlNodePtr * ((|rootNode|))
        libxml document root node buffer.
    :Return value
      :type int
        error code.
    :Note
      
=end
*/
static
int
OpenSOAPEnvelopeCreateWithLibxmlNode(/* [in]  */ xmlDocPtr	doc,
									 /* [in]  */ xmlNodePtr	envNode,
									 /* [out] */ OpenSOAPEnvelopePtr *soapEnv) {

    int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (doc && envNode && soapEnv) {
		xmlNsPtr	envNs;
		SoapVersionEnvValuesMapConstIterator i = SOAP_VERSION_ENVURI_MAP;
		for (envNs = NULL; i->soapVersion; ++i) {
			envNs = xmlSearchNsByHref(doc,
									   envNode,
									   i->soapEnvURI);
			if (envNs) {
				break;
			}
		}
		if (!i->soapVersion) {
			ret = OPENSOAP_XML_BADNAMESPACE;
		}
		else if (envNode->name && strcmp(envNode->name,
										 SOAPEnvelopeElementName) != 0) {
			/* root node name check */
			ret = OPENSOAP_XML_BADDOCUMENTTYPE;
		}
		else {
			ret = OpenSOAPEnvelopeCreateMB(i->soapVersion,
										   envNs->prefix,
										   soapEnv);
			if (OPENSOAP_SUCCEEDED(ret)) {
				/* add envelope's attribute */
				ret = OpenSOAPXMLElmAddAttrInCharEncoding(envNode,
														  (OpenSOAPXMLElmPtr)(*soapEnv));
				if (OPENSOAP_FAILED(ret)) {
					OpenSOAPEnvelopeRelease(*soapEnv);
					*soapEnv = NULL;
				}
			}
		}
	}
	
	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeSetHeaderBodyElm(soapEnv)
    create char encoding
    
    :Parameters
      :[in, out] OpenSOAPEnvelopePtr ((|soapEnv|))
        Strage Buffer of OpenSOAP Envelope Pointer
    :Return value
      :type int
        error code.
    :Note
      
=end
*/
static
int
OpenSOAPEnvelopeSetHeaderBodyElm(OpenSOAPEnvelopePtr soapEnv) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;
	
	if (soapEnv) {
		/* Envelope Node */
		OpenSOAPXMLNodePtr envNode = (OpenSOAPXMLNodePtr)soapEnv;
		/* SOAP-ENV namespace */
		OpenSOAPXMLNamespacePtr envNs = envNode->thisNamespace;
		/* Envelope child node */
		OpenSOAPXMLNodePtr envChild = envNode->children;
		/* */
		int cmpResult = 0;

		/* Header check */
		ret = envChild
			? OpenSOAPStringCompareMB(envChild->name,
									  SOAPHeaderElementName,
									  &cmpResult)
			: OPENSOAP_XML_NOHEADERBODY;
		if (OPENSOAP_SUCCEEDED(ret)) {
			if (cmpResult == 0) {
				if (envChild->thisNamespace != envNs) {
					ret = OPENSOAP_XML_BADNAMESPACE;
				}
				else {
					soapEnv->header = (OpenSOAPXMLElmPtr)envChild;
					ret = OpenSOAPBlockReplaceXMLElmChildren(soapEnv
															 ->header);
					envChild = envChild->next;
				}
			}
			if (OPENSOAP_SUCCEEDED(ret)) {
				/* Body check */
				ret = envChild
					? OpenSOAPStringCompareMB(envChild->name,
											  SOAPBodyElementName,
											  &cmpResult)
					: OPENSOAP_XML_NOHEADERBODY;
				if (cmpResult == 0) {
					if (envChild->thisNamespace != envNs) {
						ret = OPENSOAP_XML_BADNAMESPACE;
					}
					else {
						soapEnv->body = (OpenSOAPXMLElmPtr)envChild;
						ret = OpenSOAPBlockReplaceXMLElmChildren(soapEnv
																 ->body);
					}
				}
				else {
					ret = OPENSOAP_XML_NOHEADERBODY;
				}
			}
		}
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeCreateCharEncoding(chEnc, b_ary, soapEnv)
    create char encoding
    
    :Parameters
      :const cahr * [in] ((|chEnc|))
        character encoding
      :OpenSOAPByteArrayPtr [in] ((|b_ary|))
        OpenSOAP ByteArray
      :OpenSOAPEnvelopePtr * [out] ((|soapEnv|))
        Strage Buffer of OpenSOAP Envelope Pointer
    :Return value
      :type int
        error code.
    :Note
      (2001/08/28)
      b_ary̓echEnc(EUC-JP/SHIFT_JIS/UTF-8...)ɃGR[fBOB
      soapEnvɊeXp[^b_aryݒ肷B
      
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeCreateCharEncoding(const char *chEnc,
                                   OpenSOAPByteArrayPtr bAry,
                                   OpenSOAPEnvelopePtr *soapEnv) {
#if 1
    xmlDocPtr	doc = NULL;
    xmlNodePtr	envNode = NULL;

	int ret = OpenSOAPLibxmlDocCreateCharEncoding(chEnc,
												  bAry,
												  &doc,
												  &envNode);
	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPEnvelopeCreateWithLibxmlNode(doc,
												   envNode,
												   soapEnv);
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = OpenSOAPXMLElmCreateChildNodeFromLibxml((OpenSOAPXMLElmPtr)(*soapEnv),
														  envNode);
			
			if (OPENSOAP_SUCCEEDED(ret)) {
				ret = OpenSOAPEnvelopeSetHeaderBodyElm(*soapEnv);
			}
		}

		if (OPENSOAP_FAILED(ret)) {
			OpenSOAPEnvelopeRelease(*soapEnv);
			*soapEnv = NULL;
		}

		xmlFreeDoc(doc);
	}

	return ret;
#else	
    int ret = OPENSOAP_YET_IMPLEMENTATION;

#define OPENSOAP_VER_LEN	5
#define OPENSOAP_ENV_PRE	20

    xmlDocPtr	doc = NULL;
    xmlNsPtr	envNs = NULL;
    xmlNodePtr	envNode = NULL;
	xmlNodePtr	hbNode = NULL;
	xmlNodePtr	hbChildNode = NULL;
	xmlNodePtr	hbChildNextNode = NULL;
	xmlAttrPtr	nextAttr = NULL;
	xmlNsPtr	nextNs = NULL;
	int  isBlankNode;

	OpenSOAPStringPtr opensoap_nodevalue = NULL;
	OpenSOAPStringPtr attr_value = NULL;

	char soap_ver[OPENSOAP_VER_LEN];
	char env_prefix[OPENSOAP_ENV_PRE];

	OpenSOAPBlockPtr	 h_block = NULL;
	OpenSOAPBlockPtr	 b_block = NULL;

	OpenSOAPByteArrayPtr tmp_bAry = NULL;
	OpenSOAPXMLAttrPtr new_attr = NULL;

	size_t flag_addHeaderAttr; /* 0:not added; 1:added */
	size_t flag_addHeaderNs;
	size_t flag_addBodyAttr;
	size_t flag_addBodyNs;


	/* change bAry to "UTF-8" */
	ret = OpenSOAPByteArrayCreate(&tmp_bAry);
	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPStringConvertCharEncoding(chEnc,
												bAry,
												"UTF-8",
												tmp_bAry);
		if (OPENSOAP_SUCCEEDED(ret)) {
			char *tmpBuf = NULL;
			size_t tmpBufSize = 0;
			ret = OpenSOAPByteArrayRemoveXMLProcAttr(tmp_bAry,
													 &tmpBuf,
													 &tmpBufSize);
			if (OPENSOAP_SUCCEEDED(ret)) {
				doc = xmlParseMemory(tmpBuf, tmpBufSize);
				free(tmpBuf);
			}
		}
		ret = OpenSOAPByteArrayRelease(tmp_bAry);
		if (OPENSOAP_FAILED(ret)) {
			xmlFreeDoc(doc);
			return ret;
		}
	}
    
    if (doc == NULL) {
		ret = OPENSOAP_XML_BADMAKEDOCUMENT;
		return ret;
	}
	
	/* check document (root node) */
    envNode = xmlDocGetRootElement(doc);
    if (envNode == NULL) {
		xmlFreeDoc(doc);
		ret = OPENSOAP_XML_EMPTYDOCUMENT;
		return ret;
    }
	
	/* check namespace */
    envNs = xmlSearchNsByHref(doc,
							  envNode,
							  "http://schemas.xmlsoap.org/soap/envelope/");
    if (envNs == NULL) {
		envNs = xmlSearchNsByHref(doc,
								  envNode,
								  "http://www.w3.org/2001/06/soap-envelope");
		if (envNs == NULL) {
			xmlFreeDoc(doc);
			ret = OPENSOAP_XML_BADNAMESPACE;
			return ret;
		}
		else{
			strcpy(soap_ver, "1.2");
			strcpy(env_prefix, envNs->prefix);
		}
    }
	else{
		strcpy(soap_ver, "1.1");
		strcpy(env_prefix, envNs->prefix);
	}
	
	/* check document type (root node) */
    if (strcmp(envNode->name, SOAPEnvelopeElementName) != 0) {
		xmlFreeDoc(doc);
		ret = OPENSOAP_XML_BADDOCUMENTTYPE;
		return ret;
    }

	ret = OpenSOAPEnvelopeCreateMB(soap_ver, env_prefix, soapEnv);
	if (OPENSOAP_FAILED(ret)) {
		xmlFreeDoc(doc);
		return ret;
	}

	/* add envelope's attribute */
	ret = OpenSOAPXMLElmAddAttrInCharEncoding(envNode,
											  (OpenSOAPXMLElmPtr)(*soapEnv));
	if (OPENSOAP_FAILED(ret)) {
		xmlFreeDoc(doc);
		return ret;
	}

	/* add header block */
	hbNode = envNode->xmlChildrenNode;
	isBlankNode = xmlIsBlankNode(hbNode);
	if (isBlankNode == 1) {
		hbNode = hbNode->next;
	}
	if (hbNode->ns == envNs) {
		if (strcmp(hbNode->name, SOAPHeaderElementName) == 0) {
			hbChildNode = hbNode->xmlChildrenNode;
			isBlankNode = xmlIsBlankNode( hbChildNode );
			if( isBlankNode == 1 ) {
				hbChildNode = hbChildNode->next;
			}
			hbChildNextNode = hbChildNode;

			/* add header element only without header block */
			if( hbChildNextNode == NULL ){
				ret = OpenSOAPEnvelopeAddHeaderBlockMB( *soapEnv,
													    NULL,
													    &h_block );
				if( OPENSOAP_FAILED(ret) ){
					xmlFreeDoc(doc);
					return ret;
				}
				/* add header element attribute */
				ret = OpenSOAPXMLElmAddAttrInCharEncoding( hbNode, (*soapEnv)->header );
				if( OPENSOAP_FAILED(ret) ){
					xmlFreeDoc(doc);
					return ret;
				}
				/* add header element namespace */
				ret = OpenSOAPXMLElmAddNsInCharEncoding( hbNode, (*soapEnv)->header );
				if( OPENSOAP_FAILED(ret) ){
					xmlFreeDoc(doc);
					return ret;
				}
			}


			/* add header block */
			flag_addHeaderAttr = 0;
			flag_addHeaderNs = 0;
			while( hbChildNextNode != NULL ) {
				OpenSOAPByteArrayPtr mb_name = NULL;
				const unsigned char *mb_name_beg = NULL;

				ret = OpenSOAPByteArrayCreate( &mb_name );
				if( OPENSOAP_FAILED(ret) ){
					xmlFreeDoc(doc);
					return ret;
				}

				ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( hbChildNextNode->name,
																   mb_name,
																   &mb_name_beg );
				if( OPENSOAP_FAILED(ret) ){
					xmlFreeDoc(doc);
					return ret;
				}

				ret = OpenSOAPEnvelopeAddHeaderBlockMB( *soapEnv,
														mb_name_beg,
														&h_block );
				if( OPENSOAP_FAILED(ret) ) {
					xmlFreeDoc(doc);
					return ret;
				}

				ret = OpenSOAPByteArrayRelease( mb_name );
				if( OPENSOAP_FAILED(ret) ){
					xmlFreeDoc(doc);
					return ret;
				}

				/* add header element attribute */
				if( flag_addHeaderAttr == 0 ){
					ret = OpenSOAPXMLElmAddAttrInCharEncoding( hbNode, (*soapEnv)->header );
					if( OPENSOAP_FAILED(ret) ){
						xmlFreeDoc(doc);
						return ret;
					}
					flag_addHeaderAttr = 1;
				}

				/* add header element namespace */
				if( flag_addHeaderNs == 0 ){
					ret = OpenSOAPXMLElmAddNsInCharEncoding( hbNode, (*soapEnv)->header );
					if( OPENSOAP_FAILED(ret) ){
						xmlFreeDoc(doc);
						return ret;
					}
					flag_addHeaderNs = 1;
				}

				/* add header block attributes */
				ret = OpenSOAPXMLElmAddAttrInCharEncoding( hbChildNextNode, 
														   (OpenSOAPXMLElmPtr)h_block );
				if( OPENSOAP_FAILED(ret) ){
					xmlFreeDoc(doc);
					return ret;
				}

				/* add header block namespace */
				ret = OpenSOAPXMLElmAddNsInCharEncoding( hbChildNextNode,
														 (OpenSOAPXMLElmPtr)h_block );
				if( OPENSOAP_FAILED(ret) ){
					xmlFreeDoc(doc);
					return ret;
				}

				/* add header block's child node */
				if( hbChildNextNode->children != NULL ) {
					xmlNodePtr hbChildNextChild;
					hbChildNextChild = hbChildNextNode->children;
					isBlankNode = xmlIsBlankNode( hbChildNextChild );
					if( isBlankNode == 1 ){
						hbChildNextChild = hbChildNextChild->next;
					}
					if( strcmp( hbChildNextChild->name, "text" ) != 0 ){
						ret = OpenSOAPXMLElmCreateChildNodeFromLibxml( (OpenSOAPXMLElmPtr)h_block, hbChildNextNode );
						if( OPENSOAP_FAILED(ret) ) {
							xmlFreeDoc(doc);
							return ret;
						}
					}
					else{
						OpenSOAPByteArrayPtr mb_content = NULL;
						const unsigned char *mb_content_beg = NULL;
						size_t mb_content_sz = 0;
						unsigned char *mb_content_beg_noLF = NULL; 
						size_t mb_content_sz_noLF = 0;
						int i;

						ret = OpenSOAPByteArrayCreate( &mb_content );
						if( OPENSOAP_FAILED(ret) ){
							xmlFreeDoc(doc);
							return ret;
						}
						ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( hbChildNextChild->content,
																		   mb_content,
																		   &mb_content_beg );
						if( OPENSOAP_FAILED(ret) ){
							xmlFreeDoc(doc);
							return ret;
						}

						/* delete '\t' '\n' '\r' in the content */
						ret = OpenSOAPByteArrayGetBeginSizeConst(mb_content, &mb_content_beg, &mb_content_sz);
						mb_content_beg_noLF = malloc( mb_content_sz * sizeof(unsigned char) );
						if( mb_content_beg_noLF == NULL ){
							xmlFreeDoc(doc);
							ret = OPENSOAP_MEM_BADALLOC;
							return ret;
						}

						for( i=0; i<(int)mb_content_sz; i++ ){
							if( mb_content_beg[i] != '\n' && mb_content_beg[i] != '\r' && mb_content_beg[i] != '\t'){
								mb_content_beg_noLF[mb_content_sz_noLF] = mb_content_beg[i];
								mb_content_sz_noLF++;
							}
						}

						ret = OpenSOAPStringCreateWithMB( mb_content_beg_noLF, &opensoap_nodevalue );
						if( OPENSOAP_FAILED(ret) ) {
							free( mb_content_beg_noLF );
							xmlFreeDoc(doc);
							return ret;
						}
						free( mb_content_beg_noLF );

						ret = OpenSOAPBlockSetValueMB( h_block, "string", &opensoap_nodevalue );
						if( OPENSOAP_FAILED(ret) ){
							xmlFreeDoc(doc);
							return ret;
						}
						ret = OpenSOAPByteArrayRelease( mb_content );
						if( OPENSOAP_FAILED(ret) ){
							xmlFreeDoc(doc);
							return ret;
						}
					}
				}

				hbChildNextNode = hbChildNextNode->next;
				isBlankNode = xmlIsBlankNode( hbChildNextNode );
				if( isBlankNode == 1 ) {
					hbChildNextNode = hbChildNextNode->next;
				}
			}

			/* add body block after header block*/
			hbNode = hbNode->next;
			if( hbNode == NULL ){
				xmlFreeDoc(doc);
				ret = OPENSOAP_XML_NOHEADERBODY;
				return ret;
			}
			else{
				isBlankNode = xmlIsBlankNode( hbNode );
				if( isBlankNode == 1 ) {
					hbNode = hbNode->next;
				}
			}
			if (hbNode != NULL
				&& strcmp(hbNode->name, SOAPBodyElementName) == 0) {
				hbChildNode = hbNode->xmlChildrenNode;
				isBlankNode = xmlIsBlankNode( hbChildNode );
				if( isBlankNode == 1 ) {
					hbChildNode = hbChildNode->next;
				}
				hbChildNextNode = hbChildNode;

				/* add body element only without body block */
				if( hbChildNextNode == NULL ){
					ret = OpenSOAPEnvelopeAddBodyBlockMB( *soapEnv,
														  NULL,
														  &b_block );
					if( OPENSOAP_FAILED(ret) ){
						xmlFreeDoc(doc);
						return ret;
					}
					/* add body element attribute */
					ret = OpenSOAPXMLElmAddAttrInCharEncoding( hbNode, (*soapEnv)->body );
					if( OPENSOAP_FAILED(ret) ){
						xmlFreeDoc(doc);
						return ret;
					}
					/* add body element namespace */
					ret = OpenSOAPXMLElmAddNsInCharEncoding( hbNode, (*soapEnv)->body );
					if( OPENSOAP_FAILED(ret) ){
						xmlFreeDoc(doc);
						return ret;
					}
				}

				/* add body block */
				flag_addBodyAttr = 0;
				flag_addBodyNs = 0;
				while( hbChildNextNode != NULL ) {
					OpenSOAPByteArrayPtr mb_name = NULL;
					const unsigned char *mb_name_beg = NULL;

					ret = OpenSOAPByteArrayCreate( &mb_name );
					if( OPENSOAP_FAILED(ret) ){
						xmlFreeDoc(doc);
						return ret;
					}

					ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( hbChildNextNode->name,
																	   mb_name,
																	   &mb_name_beg );
					if( OPENSOAP_FAILED(ret) ){
						xmlFreeDoc(doc);
						return ret;
					}
					ret = OpenSOAPEnvelopeAddBodyBlockMB( *soapEnv,
														  mb_name_beg,
														  &b_block );
					if( OPENSOAP_FAILED(ret) ) {
						xmlFreeDoc(doc);
						return ret;
					}

					/* add body element attribute */
					if( flag_addBodyAttr == 0 ){
						ret = OpenSOAPXMLElmAddAttrInCharEncoding( hbNode, (*soapEnv)->body );
						if( OPENSOAP_FAILED(ret) ){
							xmlFreeDoc(doc);
							return ret;
						}
						flag_addBodyAttr = 1;
					}

					/* add body element namespace */
					if( flag_addBodyNs == 0 ){
						ret = OpenSOAPXMLElmAddNsInCharEncoding( hbNode, (*soapEnv)->body );
						if( OPENSOAP_FAILED(ret) ){
							xmlFreeDoc(doc);
							return ret;
						}
						flag_addBodyNs = 1;
					}

					/* add body block attributes */
					ret = OpenSOAPXMLElmAddAttrInCharEncoding( hbChildNextNode,
															   (OpenSOAPXMLElmPtr)b_block);
					if( OPENSOAP_FAILED(ret) ){
						xmlFreeDoc(doc);
						return ret;
					}

					/* add body block namespace */
					ret = OpenSOAPXMLElmAddNsInCharEncoding( hbChildNextNode,
															 (OpenSOAPXMLElmPtr)b_block);
					if( OPENSOAP_FAILED(ret) ){
						xmlFreeDoc(doc);
						return ret;
					}

					/* add body block's child node */
					if( hbChildNextNode->children != NULL ) {
						xmlNodePtr hbChildNextChild;
						hbChildNextChild = hbChildNextNode->children;
						isBlankNode = xmlIsBlankNode( hbChildNextChild );
						if( isBlankNode == 1 ){
							hbChildNextChild = hbChildNextChild->next;
						}
						if( strcmp( hbChildNextChild->name, "text" ) != 0 ){
							ret = OpenSOAPXMLElmCreateChildNodeFromLibxml( (OpenSOAPXMLElmPtr)b_block, hbChildNextNode );
							if( OPENSOAP_FAILED(ret) ) {
								xmlFreeDoc(doc);
								return ret;
							}
						}
						else{
							OpenSOAPByteArrayPtr mb_content = NULL;
							const unsigned char *mb_content_beg = NULL;
							size_t mb_content_sz = 0;
							unsigned char *mb_content_beg_noLF = NULL;
							size_t mb_content_sz_noLF = 0;
							int i;

							ret = OpenSOAPByteArrayCreate( &mb_content );
							if( OPENSOAP_FAILED(ret) ){
								xmlFreeDoc(doc);
								return ret;
							}
							ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( hbChildNextChild->content,
																			   mb_content,
																			   &mb_content_beg );
							if( OPENSOAP_FAILED(ret) ){
								xmlFreeDoc(doc);
								return ret;
							}

							/* delete '\t' '\n' '\r' in the content */
							ret = OpenSOAPByteArrayGetBeginSizeConst(mb_content, &mb_content_beg, &mb_content_sz);
							mb_content_beg_noLF = malloc( mb_content_sz * sizeof(unsigned char) );
							if( mb_content_beg_noLF == NULL ){
								ret = OPENSOAP_MEM_BADALLOC;
								xmlFreeDoc(doc);
								return ret;
							}

							for( i=0; i<(int)mb_content_sz; i++ ){
								if( mb_content_beg[i] != '\n' && mb_content_beg[i] != '\r' && mb_content_beg[i] != '\t'){
									mb_content_beg_noLF[mb_content_sz_noLF] = mb_content_beg[i];
									mb_content_sz_noLF++;
								}
							}

							ret = OpenSOAPStringCreateWithMB( mb_content_beg_noLF, &opensoap_nodevalue );
							if( OPENSOAP_FAILED(ret) ) {
								free( mb_content_beg_noLF );
								xmlFreeDoc(doc);
								return ret;
							}
							free(mb_content_beg_noLF);

							ret = OpenSOAPBlockSetValueMB( b_block, "string", &opensoap_nodevalue );
							if( OPENSOAP_FAILED(ret) ){
								xmlFreeDoc(doc);
								return ret;
							}
							ret = OpenSOAPByteArrayRelease(mb_content);
							if (OPENSOAP_FAILED(ret)) {
								xmlFreeDoc(doc);
								return ret;
							}
						}
					}

					hbChildNextNode = hbChildNextNode->next;
					isBlankNode = xmlIsBlankNode( hbChildNextNode );
					if (isBlankNode == 1) {
						hbChildNextNode = hbChildNextNode->next;
					}
				}
			}
			else {
				ret = OPENSOAP_XML_NOHEADERBODY;
				xmlFreeDoc(doc);
				return ret;
			}

		}
		/* add body block without header block */
		else if (strcmp(hbNode->name, SOAPBodyElementName) == 0) {
			hbChildNode = hbNode->xmlChildrenNode;
			isBlankNode = xmlIsBlankNode( hbChildNode );
			if( isBlankNode == 1 ) {
				hbChildNode = hbChildNode->next;
			}
			hbChildNextNode = hbChildNode;

			if( hbChildNextNode == NULL ){
				ret = OpenSOAPEnvelopeAddBodyBlockMB( *soapEnv,
													  NULL,
													  &b_block );
				if( OPENSOAP_FAILED(ret) ){
					xmlFreeDoc(doc);
					return ret;
				}
				/* add body element attribute */
				ret = OpenSOAPXMLElmAddAttrInCharEncoding( hbNode, (*soapEnv)->body );
				if( OPENSOAP_FAILED(ret) ){
					xmlFreeDoc(doc);
					return ret;
				}
				/* add body element namespace */
				ret = OpenSOAPXMLElmAddNsInCharEncoding( hbNode, (*soapEnv)->body );
				if( OPENSOAP_FAILED(ret) ){
					xmlFreeDoc(doc);
					return ret;
				}
			}

			/* add body block */
			flag_addBodyAttr = 0;
			flag_addBodyNs = 0;
			while( hbChildNextNode != NULL ) {
				OpenSOAPByteArrayPtr mb_name = NULL;
				const unsigned char *mb_name_beg = NULL;

				ret = OpenSOAPByteArrayCreate(&mb_name);
				if (OPENSOAP_FAILED(ret)) {
					OpenSOAPByteArrayRelease(mb_name);
					xmlFreeDoc(doc);
					return ret;
				}

				ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml(hbChildNextNode->name,
																  mb_name,
																  &mb_name_beg);
				if (OPENSOAP_FAILED(ret)) {
					OpenSOAPByteArrayRelease(mb_name);
					xmlFreeDoc(doc);
					return ret;
				}
				ret = OpenSOAPEnvelopeAddBodyBlockMB(*soapEnv,
													 mb_name_beg,
													 &b_block);
				if (OPENSOAP_FAILED(ret)) {
					OpenSOAPByteArrayRelease(mb_name);
					xmlFreeDoc(doc);
					return ret;
				}
				OpenSOAPByteArrayRelease(mb_name);

				/* add body element attribute */
				if( flag_addBodyAttr == 0 ){
					ret = OpenSOAPXMLElmAddAttrInCharEncoding( hbNode, (*soapEnv)->body );
					if( OPENSOAP_FAILED(ret) ){
						xmlFreeDoc(doc);
						return ret;
					}
					flag_addBodyAttr = 1;
				}
				/* add body element namespace */
				if( flag_addBodyNs == 0 ){
					ret = OpenSOAPXMLElmAddNsInCharEncoding( hbNode, (*soapEnv)->body );
					if( OPENSOAP_FAILED(ret) ){
						xmlFreeDoc(doc);
						return ret;
					}
					flag_addBodyNs = 1;
				}

				/* add body block attributes */
				ret = OpenSOAPXMLElmAddAttrInCharEncoding( hbChildNextNode,
														   (OpenSOAPXMLElmPtr)b_block);
				if( OPENSOAP_FAILED(ret) ){
					xmlFreeDoc(doc);
					return ret;
				}

				/* add body block namespace */
				ret = OpenSOAPXMLElmAddNsInCharEncoding( hbChildNextNode,
														 (OpenSOAPXMLElmPtr)b_block);
				if( OPENSOAP_FAILED(ret) ){
					xmlFreeDoc(doc);
					return ret;
				}

				/* add block's child node */
				if( hbChildNextNode->children != NULL ) {
					xmlNodePtr hbChildNextChild;
					hbChildNextChild = hbChildNextNode->children;
					isBlankNode = xmlIsBlankNode( hbChildNextChild );
					if (isBlankNode == 1) {
						hbChildNextChild = hbChildNextChild->next;
					}
					if (!hbChildNextChild || !hbChildNextChild->name
						|| strcmp(hbChildNextChild->name, "text") != 0) {
						ret = OpenSOAPXMLElmCreateChildNodeFromLibxml((OpenSOAPXMLElmPtr)b_block, hbChildNextNode);
						if (OPENSOAP_FAILED(ret)) {
							xmlFreeDoc(doc);
							return ret;
						}
					}
					else {
						OpenSOAPByteArrayPtr mb_content = NULL;
						const unsigned char *mb_content_beg = NULL;
						size_t mb_content_sz = 0;
						unsigned char *mb_content_beg_noLF = NULL;
						size_t mb_content_sz_noLF = 0;
						int i;

						ret = OpenSOAPByteArrayCreate( &mb_content );
						if( OPENSOAP_FAILED(ret) ){
							xmlFreeDoc(doc);
							return ret;
						}
						ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( hbChildNextChild->content,
																		   mb_content,
																		   &mb_content_beg );
						if( OPENSOAP_FAILED(ret) ){
							xmlFreeDoc(doc);
							return ret;
						}

						/* delete '\t' '\n' '\r' in the content */
						ret = OpenSOAPByteArrayGetBeginSizeConst(mb_content,
																 &mb_content_beg,
																 &mb_content_sz);
						mb_content_beg_noLF = malloc(mb_content_sz * sizeof(unsigned char));
						if (mb_content_beg_noLF == NULL) {
							ret = OPENSOAP_MEM_BADALLOC;
							xmlFreeDoc(doc);
							return ret;
						}

						for (i = 0; i < (int)mb_content_sz; ++i) {
							if (mb_content_beg[i] != '\n'
								&& mb_content_beg[i] != '\r'
								&& mb_content_beg[i] != '\t') {
								mb_content_beg_noLF[mb_content_sz_noLF]
									= mb_content_beg[i];
								++mb_content_sz_noLF;
							}
						}

						ret = OpenSOAPStringCreateWithMB(mb_content_beg_noLF,
														 &opensoap_nodevalue);
						if (OPENSOAP_FAILED(ret)) {
							free(mb_content_beg_noLF);
							xmlFreeDoc(doc);
							return ret;
						}
						free(mb_content_beg_noLF);

						ret = OpenSOAPBlockSetValueMB(b_block,
													  "string",
													  &opensoap_nodevalue);
						if (OPENSOAP_FAILED(ret)) {
							xmlFreeDoc(doc);
							return ret;
						}
						ret = OpenSOAPByteArrayRelease(mb_content);
						if (OPENSOAP_FAILED(ret)) {
							xmlFreeDoc(doc);
							return ret;
						}
					}
				}

				hbChildNextNode = hbChildNextNode->next;
				isBlankNode = xmlIsBlankNode(hbChildNextNode);
				if (isBlankNode == 1) {
					hbChildNextNode = hbChildNextNode->next;
				}
			}
		}
		else {
			ret = OPENSOAP_XML_NOHEADERBODY;
			xmlFreeDoc(doc);
			return ret;
		}
	}
	else {
		ret = OPENSOAP_XML_BADNAMESPACE;
		xmlFreeDoc(doc);
		return ret;
	}

	xmlFreeDoc(doc);
	ret = OPENSOAP_NO_ERROR;
	return ret;
#endif	
}


/*
=begin
--- function#OpenSOAPEnvelopeRetain(soap_env)
    SOAP Envelope t@XǉB

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
    :Return value
      :int
	    Return Code Is ErrorCode
      
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeRetain(OpenSOAPEnvelopePtr soap_env) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeRelease(soap_env)
    Release SOAP Envelope Buffer

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
    :Return value
      :int
	    Return Code Is ErrorCode
    :Note
	  SOAP Envelope쐬̈̊Js܂B
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeRelease(OpenSOAPEnvelopePtr soap_env) {
    int ret = OpenSOAPXMLElmRelease((OpenSOAPXMLElmPtr)soap_env);

    return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeAddHeaderBlockString(soap_env, block_name, h_block)
    Add SOAP Header Block(OpenSOAPString)

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :OpenSOAPStringPtr [in] ((|block_name|))
        SOAP Block Name
      :OpenSOAPBlockPtr * [out] ((|h_block|))
        OpenSOAP Header Block
    :Return value
      :int
	    Return Code Is ErrorCode
    :Note
      SOAP EnvelopëHeader Blockǉ܂B
      (2001/10/11)
      This Function Is Static Function
  
=end
*/
/* extern */
static
int
/* OPENSOAP_API */
OpenSOAPEnvelopeAddHeaderBlockString(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                                 OpenSOAPStringPtr /* [in] */ block_name,
                                 OpenSOAPBlockPtr * /* [out] */ h_block) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLElmPtr second_b;

	if( !soap_env || !h_block ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( !soap_env->header && block_name == NULL ){
		/* add body element only ( no child---body block )*/
		ret = OpenSOAPXMLElmCreate( &soap_env->header );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringCreate( &(soap_env->header->super.name) );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringCreateWithMB(SOAPHeaderElementName,
										 &soap_env->header->super.name);
		if(ret == OPENSOAP_PARAMETER_BADVALUE){
			return ret;
		}

		if( soap_env->envElm.super.children == NULL ){
			soap_env->envElm.super.children = (OpenSOAPXMLNodePtr)soap_env->header;
		}
		else{
			second_b = (OpenSOAPXMLElmPtr)soap_env->envElm.super.children;
			soap_env->envElm.super.children = (OpenSOAPXMLNodePtr)soap_env->header;
			soap_env->header->super.next = (OpenSOAPXMLNodePtr)second_b;
		}

		soap_env->header->super.parent = (OpenSOAPXMLNodePtr)soap_env;

		*h_block = NULL;

		ret = OPENSOAP_NO_ERROR;
		return ret;
	}

	if( !soap_env->header && block_name != NULL ){
		ret = OpenSOAPXMLElmCreate( &soap_env->header );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringCreate( &(soap_env->header->super.name) );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringCreateWithMB(SOAPHeaderElementName,
										 &soap_env->header->super.name);
		if(ret == OPENSOAP_PARAMETER_BADVALUE){
			return ret;
		}

		
		if( soap_env->envElm.super.children == NULL ){
			soap_env->envElm.super.children = (OpenSOAPXMLNodePtr)soap_env->header;
		}
		else{
			second_b = (OpenSOAPXMLElmPtr)soap_env->envElm.super.children;
			soap_env->envElm.super.children = (OpenSOAPXMLNodePtr)soap_env->header;
			soap_env->header->super.next = (OpenSOAPXMLNodePtr)second_b;
		}

		soap_env->header->super.parent = (OpenSOAPXMLNodePtr)soap_env;
	}

	ret = OpenSOAPBlockCreate( h_block );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	ret = OpenSOAPXMLElmAddChildString( soap_env->header, block_name, (OpenSOAPXMLElmPtr*)h_block );

    return ret;
}


/*
=begin
--- function#OpenSOAPEnvelopeAddHeaderBlockMB(soap_env, block_name, h_block)
    Add SOAP Header Block(MB)

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :const char * [in] ((|block_name|))
        SOAP Block Name
      :OpenSOAPBlockPtr * [out] ((|h_block|))
        OpenSOAP Header Block
    :Return value
      :int
	    Return Code Is ErrorCode
    :Note
      SOAP EnvelopëHeader Blockǉ܂B
  
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeAddHeaderBlockMB(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                                 const char * /* [in] */ block_name,
                                 OpenSOAPBlockPtr * /* [out] */ h_block) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLElmPtr second_b;

	if( !soap_env || !h_block ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( !soap_env->header && block_name == NULL ){
		/* add body element only ( no child---body block )*/
		ret = OpenSOAPXMLElmCreate( &soap_env->header );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringCreate( &(soap_env->header->super.name) );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringSetStringMB(soap_env->header->super.name,
										SOAPHeaderElementName);
		if(ret == OPENSOAP_PARAMETER_BADVALUE){
			return ret;
		}

		if( soap_env->envElm.super.children == NULL ){
			soap_env->envElm.super.children = (OpenSOAPXMLNodePtr)soap_env->header;
		}
		else{
			second_b = (OpenSOAPXMLElmPtr)soap_env->envElm.super.children;
			soap_env->envElm.super.children = (OpenSOAPXMLNodePtr)soap_env->header;
			soap_env->header->super.next = (OpenSOAPXMLNodePtr)second_b;
		}

		soap_env->header->super.parent = (OpenSOAPXMLNodePtr)soap_env;
		soap_env->header->super.thisNamespace = soap_env->envElm.super.thisNamespace;

		*h_block = NULL;

		ret = OPENSOAP_NO_ERROR;
		return ret;
	}

	if( !soap_env->header && block_name != NULL ){
		ret = OpenSOAPXMLElmCreate( &soap_env->header );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringCreate( &(soap_env->header->super.name) );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringSetStringMB(soap_env->header->super.name,
										SOAPHeaderElementName);
		if(ret == OPENSOAP_PARAMETER_BADVALUE){
			return ret;
		}

		
		if( soap_env->envElm.super.children == NULL ){
			soap_env->envElm.super.children = (OpenSOAPXMLNodePtr)soap_env->header;
		}
		else{
			second_b = (OpenSOAPXMLElmPtr)soap_env->envElm.super.children;
			soap_env->envElm.super.children = (OpenSOAPXMLNodePtr)soap_env->header;
			soap_env->header->super.next = (OpenSOAPXMLNodePtr)second_b;
		}

		soap_env->header->super.parent = (OpenSOAPXMLNodePtr)soap_env;
		soap_env->header->super.thisNamespace = soap_env->envElm.super.thisNamespace;
	}

	ret = OpenSOAPBlockCreate( h_block );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	ret = OpenSOAPXMLElmAddChildMB( soap_env->header, block_name, (OpenSOAPXMLElmPtr*)h_block );

    return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeAddHeaderBlockWC(soap_env, block_name, h_block)
    Add SOAP Header Block(WC)

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :const wchar_t * [in] ((|block_name|))
        SOAP Block Name
      :OpenSOAPBlockPtr * [out] ((|h_block|))
        OpenSOAP Header Block
    :Return value
      :int
	    Return Code Is ErrorCode
    :Note
      SOAP EnvelopëHeader Blockǉ܂B
      
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeAddHeaderBlockWC(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                                 const wchar_t * /* [in] */ block_name,
                                 OpenSOAPBlockPtr * /* [out] */ h_block) {
	OpenSOAPStringPtr block_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(block_name, &block_name_str);
	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPEnvelopeAddHeaderBlockString(soap_env, block_name_str, h_block);
		OpenSOAPStringRelease(block_name_str);
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeGetNextHeaderBlock(soap_env, h_block)
    Get SOAP Header Block

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :OpenSOAPBlockPtr * [in, out] ((|h_block|))
        if OpenSOAP Header Block. *h_block as in NULL then
        Return to First Header Block
    :Return value
      :int
	    Return Code Is ErrorCode
      
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeGetNextHeaderBlock(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                                   OpenSOAPBlockPtr * /* [in, out] */ h_block) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !soap_env || !h_block ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( soap_env->header != NULL ){
		ret = OpenSOAPXMLElmGetNextChild( soap_env->header,
										  (OpenSOAPXMLElmPtr *)(h_block) );
	}
	else{
		*h_block = NULL;
		ret = OPENSOAP_NO_ERROR;
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeGetHeaderBlockString(soap_env, block_name, h_block)
    Get SOAP Header Block(OpenSOAPString)

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :OpenSOAPStringPtr [in] ((|block_name|))
        SOAP Header Block name.
      :OpenSOAPBlockPtr * [out] ((|h_block|))
        Header Block return buffer's pointer.
    :Return value
      :int
	    Return Code Is ErrorCode
	:Note
	  This Function Is Static Function
      
=end
*/
/* extern */
static
int
/* OPENSOAP_API */
OpenSOAPEnvelopeGetHeaderBlockString(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                               OpenSOAPStringPtr /* [in] */ block_name,
                               OpenSOAPBlockPtr * /* [out] */ h_block) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !soap_env || !h_block ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	ret = OpenSOAPXMLElmGetChildString( soap_env->header, 
									block_name, 
									(OpenSOAPXMLElmPtr*)h_block );

    return ret;
}


/*
=begin
--- function#OpenSOAPEnvelopeGetHeaderBlockMB(soap_env, block_name, h_block)
    Get SOAP Header Block(MB)

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :const char * [in] ((|block_name|))
        SOAP Header Block name.
      :OpenSOAPBlockPtr * [out] ((|h_block|))
        Header Block return buffer's pointer.
    :Return value
      :int
	    Return Code Is ErrorCode
      
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeGetHeaderBlockMB(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                               const char * /* [in] */ block_name,
                               OpenSOAPBlockPtr * /* [out] */ h_block) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !soap_env || !h_block ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	ret = OpenSOAPXMLElmGetChildMB( soap_env->header, 
									block_name, 
									(OpenSOAPXMLElmPtr*)h_block );

    return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeGetHeaderBlockWC(soap_env, block_name, h_block)
    Get SOAP Header Block(WC)

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :const wchar_t * [in] ((|block_name|))
        SOAP Header Block name.
      :OpenSOAPBlockPtr * [out] ((|h_block|))
        Header Block return buffer's pointer.
        
    :Return value
      :int
	    Return Code Is ErrorCode
      
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeGetHeaderBlockWC(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                               const wchar_t * /* [in] */ block_name,
                               OpenSOAPBlockPtr * /* [out] */ h_block) {
	OpenSOAPStringPtr block_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(block_name, &block_name_str);
	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPEnvelopeGetHeaderBlockString(soap_env, block_name_str, h_block);
		OpenSOAPStringRelease(block_name_str);
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeAddBodyBlockString(soap_env, block_name, b_block)
    Add SOAP Body Block(OpenSOAPString)

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :OpenSOAPStringPtr [in] ((|block_name|))
        SOAP Block Name
		If block_name==NULL, add body element only (no body block)
      :OpenSOAPBlockPtr * [out] ((|b_block|))
        OpenSOAP Body Block
        
    :Return value
      :int
	    Return Code Is ErrorCode
	:Note
	  This Function Is Static Function
      
=end
*/
/* extern */
static
int
/* OPENSOAP_API */
OpenSOAPEnvelopeAddBodyBlockString(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                                 OpenSOAPStringPtr /* [in] */ block_name,
                                 OpenSOAPBlockPtr * /* [out] */ b_block) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !soap_env || !b_block ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( !soap_env->body && block_name == NULL ){
		/* add body element only ( no child---body block )*/
		ret = OpenSOAPXMLElmCreate( &soap_env->body );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringCreate( &(soap_env->body->super.name) );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringSetStringMB(soap_env->body->super.name,
										SOAPBodyElementName);
		if(ret == OPENSOAP_PARAMETER_BADVALUE){
			return ret;
		}

		if( soap_env->envElm.super.children == NULL ){
			soap_env->envElm.super.children = (OpenSOAPXMLNodePtr)soap_env->body;
		}
		else{
			soap_env->envElm.super.children->next = (OpenSOAPXMLNodePtr)soap_env->body;
			soap_env->body->super.prev = soap_env->envElm.super.children;
		}

		soap_env->body->super.parent = (OpenSOAPXMLNodePtr)soap_env;

		*b_block = NULL;
		
		ret = OPENSOAP_NO_ERROR;
		return ret;
	}
	
	if( !soap_env->body && block_name != NULL ){
		/* add body element and its child---body block named by block_name */
		ret = OpenSOAPXMLElmCreate( &soap_env->body );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringCreate( &(soap_env->body->super.name) );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringSetStringMB(soap_env->body->super.name,
										SOAPBodyElementName);
		if(ret == OPENSOAP_PARAMETER_BADVALUE){
			return ret;
		}

		if( soap_env->envElm.super.children == NULL ){
			soap_env->envElm.super.children = (OpenSOAPXMLNodePtr)soap_env->body;
		}
		else{
			soap_env->envElm.super.children->next = (OpenSOAPXMLNodePtr)soap_env->body;
			soap_env->body->super.prev = soap_env->envElm.super.children;
		}
		soap_env->body->super.parent = (OpenSOAPXMLNodePtr)soap_env;
	}


	ret = OpenSOAPBlockCreate( b_block );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	ret = OpenSOAPXMLElmAddChildString( soap_env->body, block_name, (OpenSOAPXMLElmPtr*)b_block );

    return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeAddBodyBlockMB(soap_env, block_name, b_block)
    Add SOAP Body Block(MB)

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :const char * [in] ((|block_name|))
        SOAP Block Name
		If block_name==NULL, add body element only (no body block)
      :OpenSOAPBlockPtr * [out] ((|b_block|))
        OpenSOAP Body Block
        
    :Return value
      :int
	    Return Code Is ErrorCode
      
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeAddBodyBlockMB(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                                 const char * /* [in] */ block_name,
                                 OpenSOAPBlockPtr * /* [out] */ b_block) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !soap_env || !b_block ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( !soap_env->body && block_name == NULL ){
		/* add body element only ( no child---body block )*/
		ret = OpenSOAPXMLElmCreate( &soap_env->body );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringCreate( &(soap_env->body->super.name) );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringSetStringMB(soap_env->body->super.name,
										SOAPBodyElementName);
		if(ret == OPENSOAP_PARAMETER_BADVALUE){
			return ret;
		}

		if( soap_env->envElm.super.children == NULL ){
			soap_env->envElm.super.children = (OpenSOAPXMLNodePtr)soap_env->body;
		}
		else{
			soap_env->envElm.super.children->next = (OpenSOAPXMLNodePtr)soap_env->body;
			soap_env->body->super.prev = soap_env->envElm.super.children;
		}

		soap_env->body->super.parent = (OpenSOAPXMLNodePtr)soap_env;
		soap_env->body->super.thisNamespace = soap_env->envElm.super.thisNamespace;

		*b_block = NULL;
		
		ret = OPENSOAP_NO_ERROR;
		return ret;
	}
	
	if( !soap_env->body && block_name != NULL ){
		/* add body element and its child---body block named by block_name */
		ret = OpenSOAPXMLElmCreate( &soap_env->body );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringCreate( &(soap_env->body->super.name) );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPStringSetStringMB(soap_env->body->super.name,
										SOAPBodyElementName);
		if(ret == OPENSOAP_PARAMETER_BADVALUE){
			return ret;
		}

		if( soap_env->envElm.super.children == NULL ){
			soap_env->envElm.super.children = (OpenSOAPXMLNodePtr)soap_env->body;
		}
		else{
			soap_env->envElm.super.children->next = (OpenSOAPXMLNodePtr)soap_env->body;
			soap_env->body->super.prev = soap_env->envElm.super.children;
		}
		soap_env->body->super.parent = (OpenSOAPXMLNodePtr)soap_env;
		soap_env->body->super.thisNamespace = soap_env->envElm.super.thisNamespace;
	}


	ret = OpenSOAPBlockCreate( b_block );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	ret = OpenSOAPXMLElmAddChildMB( soap_env->body, block_name, (OpenSOAPXMLElmPtr*)b_block );

    return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeAddBodyBlockWC(soap_env, block_name, b_block)
    Add SOAP Body Block(WC)

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :const wchar_t * [in] ((|block_name|))
        SOAP Block Name
      :OpenSOAPBlockPtr * [out] ((|b_block|))
        OpenSOAP Body Block
        
    :Return value
      :int
	    Return Code Is ErrorCode
      
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeAddBodyBlockWC(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                                 const wchar_t * /* [in] */ block_name,
                                 OpenSOAPBlockPtr * /* [out] */ b_block) {
	OpenSOAPStringPtr block_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(block_name, &block_name_str);
	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPEnvelopeAddBodyBlockString(soap_env, block_name_str, b_block);
		OpenSOAPStringRelease(block_name_str);
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeGetNextBodyBlock(soap_env, b_block)
    Get SOAP Body Block

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :OpenSOAPBlockPtr * [in, out] ((|b_block|))
        if OpenSOAP Body Block. *b_block as in NULL then
        Return to First Body Block
    :Return value
      :int
	    Return Code Is ErrorCode
      
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeGetNextBodyBlock(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                                   OpenSOAPBlockPtr * /* [in, out] */ b_block) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !soap_env || !b_block ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( soap_env->body != NULL ){
		ret = OpenSOAPXMLElmGetNextChild( soap_env->body,
										  (OpenSOAPXMLElmPtr *)(b_block) );
	}
	else{
		*b_block = NULL;
		ret = OPENSOAP_NO_ERROR;
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeGetBodyBlockString(soap_env, block_name, b_block)
    Get SOAP Body Block(OpenSOAPString)

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :OpenSOAPStringPtr [in] ((|block_name|))
        SOAP Body Block name.
      :OpenSOAPBlockPtr * [out] ((|b_block|))
        Body Block return buffer's pointer.
        
    :Return value
      :int
	    Return Code Is ErrorCode

    :Note
      (2001/08/28)
      block_nameb_block߂B
	  (2001/10/11)
	  This Function Is Static Function
=end
*/
/* extern */
static
int
/* OPENSOAP_API */
OpenSOAPEnvelopeGetBodyBlockString(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                               OpenSOAPStringPtr /* [in] */ block_name,
                               OpenSOAPBlockPtr * /* [out] */ b_block) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !soap_env || !b_block ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	ret = OpenSOAPXMLElmGetChildString( soap_env->body, 
								  block_name, 
								 (OpenSOAPXMLElmPtr*)b_block );

    return ret;
}


/*
=begin
--- function#OpenSOAPEnvelopeGetBodyBlockMB(soap_env, block_name, b_block)
    Get SOAP Body Block(MB)

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :const char * [in] ((|block_name|))
        SOAP Body Block name.
      :OpenSOAPBlockPtr * [out] ((|b_block|))
        Body Block return buffer's pointer.
        
    :Return value
      :int
	    Return Code Is ErrorCode

    :Note
      (2001/08/28)
      block_nameb_block߂B
      
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeGetBodyBlockMB(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                               const char * /* [in] */ block_name,
                               OpenSOAPBlockPtr * /* [out] */ b_block) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !soap_env || !b_block ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	ret = OpenSOAPXMLElmGetChildMB( soap_env->body, 
									block_name, 
									(OpenSOAPXMLElmPtr*)b_block );

    return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeGetBodyBlockWC(soap_env, block_name, b_block)
    Get SOAP Body Block(WC)

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP EnvelopeB
      :const wchar_t * [in] ((|block_name|))
        SOAP Body Block name.
      :OpenSOAPBlockPtr * [out] ((|b_block|))
        Body Block return buffer's pointer.
        
    :Return value
      :int
	    Return Code Is ErrorCode

    :Note
      (2001/08/28)
      block_nameb_block߂B
      
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeGetBodyBlockWC(OpenSOAPEnvelopePtr /* [in] */ soap_env,
                               const wchar_t * /* [in] */ block_name,
                               OpenSOAPBlockPtr * /* [out] */ b_block) {
	OpenSOAPStringPtr block_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(block_name, &block_name_str);
	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPEnvelopeGetBodyBlockString(soap_env, block_name_str, b_block);
		OpenSOAPStringRelease(block_name_str);
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPByteArrayAddXMLProc(bAry, chEnc)
    add XML Proc line.

    :Parameters
      :[in,out] OpenSOAPByteArrayPtr ((|bAry|))
        UTF-8 encode XML Element serialized data buffer.
      :[in]     const char  * ((|chEnc|))
        character encoding. i.e. "EUC-JP", "SHIFT-JIS", "UTF-8"
    :Return value
      :int
	    error code
=end
 */
static
int
OpenSOAPByteArrayAddXMLProc(/* [in,out] */ OpenSOAPByteArrayPtr bAry,
							/* [in]     */ const char *chEnc) {
	static const char XML_PROC_PRE[]
		= "<?xml version=\"1.0\" encoding=\"";
	static const char XML_PROC_POST[]
		= "\"?>\n";
	static const size_t XML_PROC_PRE_SIZE
		= sizeof(XML_PROC_PRE) - 1;
	static const size_t XML_PROC_POST_SIZE
		= sizeof(XML_PROC_POST) - 1;
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (bAry && chEnc) {
		const unsigned char *bAryBeg = NULL;
		size_t bArySize = 0;
		/* get XML element serialize data address and size */
		ret = OpenSOAPByteArrayGetBeginSizeConst(bAry,
												 &bAryBeg,
												 &bArySize);
		if (OPENSOAP_SUCCEEDED(ret)) {
			size_t chEncSize = strlen(chEnc);
			size_t bufSize
				= XML_PROC_PRE_SIZE
				+ chEncSize
				+ XML_PROC_POST_SIZE
				+ bArySize;
			char *buf = malloc(bufSize);
			if (!buf) {
				ret = OPENSOAP_MEM_BADALLOC;
			}
			else {
				char *i = buf;
				/* add XML Proc prefix for encoding value */
				memcpy(i, XML_PROC_PRE, XML_PROC_PRE_SIZE);
				i += XML_PROC_PRE_SIZE;
				/* add encoding value */
				memcpy(i, chEnc, chEncSize);
				i += chEncSize;
				/* add XML Proc postfix for encoding value */
				memcpy(i, XML_PROC_POST, XML_PROC_POST_SIZE);
				i += XML_PROC_POST_SIZE;
				/* add xml element serialize value */
				memcpy(i, bAryBeg, bArySize);

				/* set XML Proc line add data */
				ret = OpenSOAPByteArraySetData(bAry,
											   buf,
											   bufSize);
				
				
				free(buf);
			}
			
		}
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeGetCharEncodingString(soapEnv, chEnc, bAry)
    SOAP Envelope character encoding output.

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soapEnv|))
        OpenSOAP EnvelopeB
      :const char  * [in] ((|chEnc|))
        character encoding. i.e. "EUC-JP", "SHIFT-JIS", "UTF-8"
      :OpenSOAPByteArrayPtr [out] ((|bAry|))
        Strage Buffer
    :Return value
      :int
	    Return Code Is ErrorCode
 
    :Note
      (2001/08/28)
      soapEnvSOAP MessagegݗĂB
     
=end
 */
extern
int
OPENSOAP_API
OpenSOAPEnvelopeGetCharEncodingString(/* [in]  */ OpenSOAPEnvelopePtr soapEnv,
                                      /* [in]  */ const char *chEnc,
                                      /* [out] */ OpenSOAPByteArrayPtr bAry) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;
	static const char MIDDLE_CHARSET[] = "UTF-8";

	if (soapEnv && bAry) {
		/* get current codeset */
		ret = chEnc ? OPENSOAP_NO_ERROR
			: OpenSOAPLocaleGetCurrentCodeset(&chEnc);
		if (OPENSOAP_SUCCEEDED(ret)) {
			OpenSOAPByteArrayPtr xmlElmBAry = NULL;
			/* create xml element serialize buffer */
			ret = OpenSOAPByteArrayCreate(&xmlElmBAry);
			if (OPENSOAP_SUCCEEDED(ret)) {
				/* get xml element serialize data */
				ret = OpenSOAPXMLElmGetCharEncodingString((OpenSOAPXMLElmPtr)soapEnv,
														  MIDDLE_CHARSET,
														  xmlElmBAry);
				if (OPENSOAP_SUCCEEDED(ret)) {
					/* add XML Proc line */
					ret = OpenSOAPByteArrayAddXMLProc(xmlElmBAry,
													  chEnc);
					if (OPENSOAP_SUCCEEDED(ret)) {
						/* convert char-encoding */
						ret = OpenSOAPStringConvertCharEncoding(MIDDLE_CHARSET,
																xmlElmBAry,
																chEnc,
																bAry);
					}
				}
				/* release temporary buffer */
				OpenSOAPByteArrayRelease(xmlElmBAry);
			}
		}
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeGetHeaderCharEncodingString(soap_env, ch_enc, b_ary)
    Soap Header character encoding output

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP Envelope.
      :const char  * [in] ((|ch_enc|))
        character encoding. i.e. "EUC-JP", "SHIFT-JIS", "UTF-8"
      :OpenSOAPByteArrayPtr [out] ((|b_ary|))
        Result Buffer        
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeGetHeaderCharEncodingString(OpenSOAPEnvelopePtr /* [in] */ soap_env,
										    const char * /* [in] */ ch_enc,
										    OpenSOAPByteArrayPtr /* [out] */ b_ary) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !soap_env || !soap_env->header || !b_ary ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
	}
	else{
		ret = OpenSOAPXMLElmGetCharEncodingString( soap_env->header,
												   ch_enc,
												   b_ary );
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeGetBodyCharEncodingString(soap_env, ch_enc, b_ary)
    Soap Body character encoding output

    :Parameters
      :OpenSOAPEnvelopePtr [in] ((|soap_env|))
        OpenSOAP Envelope.
      :const char  * [in] ((|ch_enc|))
        character encoding. i.e. "EUC-JP", "SHIFT-JIS", "UTF-8"
      :OpenSOAPByteArrayPtr [out] ((|b_ary|))
        Result Buffer        
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeGetBodyCharEncodingString(/* [in]  */ OpenSOAPEnvelopePtr soap_env,
										  /* [in]  */ const char *ch_enc,
										  /* [out] */ OpenSOAPByteArrayPtr b_ary) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !soap_env || !soap_env->body || !b_ary ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
	}
	else {
		ret = OpenSOAPXMLElmGetCharEncodingString(soap_env->body,
												  ch_enc,
												  b_ary);
	}

    return ret;
}

/*
 */
static
int
OpenSOAPEnvelopeAddFaultBlock(/* [out] */ OpenSOAPEnvelopePtr soapEnv,
							  /* [out] */ OpenSOAPBlockPtr *faultBlock) {
	static const char FAULT_BLOCK_NAME[] = "Fault";
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (soapEnv && faultBlock) {
		ret = OpenSOAPEnvelopeAddBodyBlockMB(soapEnv,
											 FAULT_BLOCK_NAME,
											 faultBlock);
		if (OPENSOAP_SUCCEEDED(ret)) {
			/* */
			((OpenSOAPXMLNodePtr)*faultBlock)->thisNamespace
				=	((OpenSOAPXMLNodePtr)soapEnv)->thisNamespace;
		}
	}

	return ret;
}

/*
 */
static
int
OpenSOAPBlockSetFaultValuesString(/* [out] */ OpenSOAPBlockPtr faultBlock,
								  /* [in, out]  */ OpenSOAPStringPtr faultCode,
								  /* [in, out]  */
								  OpenSOAPStringPtr faultString,
								  /* [in]  */ int isValueDup) {
	static const wchar_t FAULT_CODE_NAME[] = L"faultcode";
	static const wchar_t FAULT_STRING_NAME[] = L"faultstring";
	int ret = OPENSOAP_PARAMETER_BADVALUE;
	if (faultBlock && faultCode && faultString) {
		OpenSOAPStringPtr faultCodeName = NULL;
		ret = OpenSOAPStringCreateWithWC(FAULT_CODE_NAME,
										 &faultCodeName);
		if (OPENSOAP_SUCCEEDED(ret)) {
			OpenSOAPStringPtr faultStringName = NULL;
			ret = OpenSOAPStringCreateWithWC(FAULT_STRING_NAME,
											 &faultStringName);
			if (OPENSOAP_SUCCEEDED(ret)) {
				OpenSOAPXMLElmPtr faultElm = (OpenSOAPXMLElmPtr)faultBlock;
				ret = OpenSOAPXMLElmSetChildValueAsString(faultElm,
														  faultCodeName,
														  isValueDup,
														  faultCode);
				if (OPENSOAP_SUCCEEDED(ret)) {
					ret = OpenSOAPXMLElmSetChildValueAsString(faultElm,
															  faultStringName,
															  isValueDup,
															  faultString);
					if (OPENSOAP_FAILED(ret)) {
						OpenSOAPXMLElmRemoveChildString(faultElm,
														faultStringName,
														isValueDup,
														&faultString);
					}
				}
				if (OPENSOAP_FAILED(ret)) {
					OpenSOAPXMLElmRemoveChildString(faultElm,
													faultCodeName,
													isValueDup,
													&faultString);
				}

				OpenSOAPStringRelease(faultStringName);
			}

			OpenSOAPStringRelease(faultCodeName);
		}
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeAddFaultString(soapEnv, faultCode, faultString, isValueDup, faultBlock)
    OpenSOAP Envelope instance create(MB)

    :Parameters
      :[out] OpenSOAPEnvelopePtr ((|soapEnv|))
        OpenSOAP Envelope Pointer
      :[in]  OpenSOAPStringPtr  ((|faultCode|))
        SOAP Fault's faultcode.
      :[in]  OpenSOAPStringPtr  ((|faultString|))
        SOAP Fault's faultstring.
      :[in]  int  ((|isValueDup|))
        faultCode and faultString duplicate flag.
      :[out] OpenSOAPBlockPtr * ((|faultBlock|))
        Strage Buffer of OpenSOAP Fault Block Pointer
    :Return value
      :int 
	    error code.
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeAddFaultString(/* [out] */ OpenSOAPEnvelopePtr soapEnv,
							   /* [in]  */ OpenSOAPStringPtr faultCode,
							   /* [in]  */ OpenSOAPStringPtr faultString,
							   /* [in]  */ int isValueDup,
							   /* [out] */ OpenSOAPBlockPtr *faultBlock) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (faultCode && faultString) {
		ret = OpenSOAPEnvelopeAddFaultBlock(soapEnv, faultBlock);
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = OpenSOAPBlockSetFaultValuesString(*faultBlock,
													faultCode,
													faultString,
													isValueDup);
		}
	}
	
	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeAddFaultMB(soapEnv, faultCode, faultString, faultBlock)
    OpenSOAP Envelope instance create(MB)

    :Parameters
      :[out] OpenSOAPEnvelopePtr ((|soapEnv|))
        OpenSOAP Envelope Pointer
      :[in]  const char * ((|faultCode|))
        SOAP Fault's faultcode.
      :[in]  const char * ((|faultString|))
        SOAP Fault's faultstring.
      :[out] OpenSOAPBlockPtr * ((|faultBlock|))
        Strage Buffer of OpenSOAP Fault Block Pointer
    :Return value
      :int 
	    error code.
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeAddFaultMB(/* [out] */ OpenSOAPEnvelopePtr soapEnv,
						   /* [in]  */ const char *faultCode,
						   /* [in]  */ const char *faultString,
						   /* [out] */ OpenSOAPBlockPtr *faultBlock) {
	OpenSOAPStringPtr faultCodeStr = NULL;
	OpenSOAPStringPtr faultStringStr = NULL;
	int ret = OpenSOAPStringCreateWithMB(faultCode, &faultCodeStr);

	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPStringCreateWithMB(faultString, &faultStringStr);
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = OpenSOAPEnvelopeAddFaultString(soapEnv,
												 faultCodeStr,
												 faultStringStr,
												 0,
												 faultBlock);
		}
		else {
			OpenSOAPStringRelease(faultCodeStr);
		}
	}
	
	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeAddFaultWC(soapEnv, faultCode, faultString, faultBlock)
    OpenSOAP Envelope instance create(WC)

    :Parameters
      :[out] OpenSOAPEnvelopePtr ((|soapEnv|))
        OpenSOAP Envelope Pointer
      :[in]  const wchar_t * ((|faultCode|))
        SOAP Fault's faultcode.
      :[in]  const wchar_t * ((|faultString|))
        SOAP Fault's faultstring.
      :[out] OpenSOAPBlockPtr * ((|faultBlock|))
        Strage Buffer of OpenSOAP Fault Block Pointer
    :Return value
      :int 
	    error code.
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeAddFaultWC(/* [out] */ OpenSOAPEnvelopePtr soapEnv,
						   /* [in]  */ const wchar_t *faultCode,
						   /* [in]  */ const wchar_t *faultString,
						   /* [out] */ OpenSOAPBlockPtr *faultBlock) {
	OpenSOAPStringPtr faultCodeStr = NULL;
	OpenSOAPStringPtr faultStringStr = NULL;
	int ret = OpenSOAPStringCreateWithWC(faultCode, &faultCodeStr);

	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPStringCreateWithWC(faultString, &faultStringStr);
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = OpenSOAPEnvelopeAddFaultString(soapEnv,
												 faultCodeStr,
												 faultStringStr,
												 0,
												 faultBlock);
		}
		else {
			OpenSOAPStringRelease(faultCodeStr);
		}
	}
	
	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeCreateMB(soapVer, envPrefix, faultCode, faultString, faultBlock, soapEnv)
    OpenSOAP Envelope instance create(MB)

    :Parameters
      :[in]  const cahr * ((||soapVer|))
        SOAP Virsion
      :[in]  const char * ((|envPrefix|))
        SOAP Envelope namespace prefix.
      :[in]  const char * ((|faultCode|))
        SOAP Fault's faultcode.
      :[in]  const char * ((|faultString|))
        SOAP Fault's faultstring.
      :[out] OpenSOAPBlockPtr * ((|faultBlock|))
        Strage Buffer of OpenSOAP Fault Block Pointer
      :[out] OpenSOAPEnvelopePtr * ((|soapEnv|))
        Strage Buffer of OpenSOAP Envelope Pointer
    :Return value
      :int 
	    error code.
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeCreateFaultMB(/* [in]  */ const char *soapVer,
							  /* [in]  */ const char *envPrefix,
							  /* [in]  */ const char *faultCode,
							  /* [in]  */ const char *faultString,
							  /* [out] */ OpenSOAPBlockPtr *faultBlock,
							  /* [out] */ OpenSOAPEnvelopePtr *soapEnv) {
	int ret = OpenSOAPEnvelopeCreateMB(soapVer,
									   envPrefix,
									   soapEnv);
	if (OPENSOAP_SUCCEEDED(ret)) {
		OpenSOAPBlockPtr dummyBlock = NULL;
		if (!faultBlock) {
			faultBlock = &dummyBlock;
		}
		ret = OpenSOAPEnvelopeAddFaultMB(*soapEnv,
										 faultCode,
										 faultString,
										 faultBlock);
		if (OPENSOAP_FAILED(ret)) {
			OpenSOAPEnvelopeRelease(*soapEnv);
			*soapEnv = NULL;
		}
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPEnvelopeCreateMB(soapVer, envPrefix, faultCode, faultString, faultBlock, soapEnv)
    OpenSOAP Envelope instance create(MB)

    :Parameters
      :[in]  const cahr * ((||soapVer|))
        SOAP Virsion
      :[in]  const char * ((|envPrefix|))
        SOAP Envelope namespace prefix.
      :[in]  const char * ((|faultCode|))
        SOAP Fault's faultcode.
      :[in]  const char * ((|faultString|))
        SOAP Fault's faultstring.
      :[out] OpenSOAPBlockPtr * ((|faultBlock|))
        Strage Buffer of OpenSOAP Fault Block Pointer
      :[out] OpenSOAPEnvelopePtr * ((|soapEnv|))
        Strage Buffer of OpenSOAP Envelope Pointer
    :Return value
      :int 
	    error code.
=end
*/
extern
int
OPENSOAP_API
OpenSOAPEnvelopeCreateFaultWC(/* [in]  */ const wchar_t *soapVer,
							  /* [in]  */ const wchar_t *envPrefix,
							  /* [in]  */ const wchar_t *faultCode,
							  /* [in]  */ const wchar_t *faultString,
							  /* [out] */ OpenSOAPBlockPtr *faultBlock,
							  /* [out] */ OpenSOAPEnvelopePtr *soapEnv) {
	int ret = OpenSOAPEnvelopeCreateWC(soapVer,
									   envPrefix,
									   soapEnv);
	if (OPENSOAP_SUCCEEDED(ret)) {
		OpenSOAPBlockPtr dummyBlock = NULL;
		if (!faultBlock) {
			faultBlock = &dummyBlock;
		}
		ret = OpenSOAPEnvelopeAddFaultWC(*soapEnv,
										 faultCode,
										 faultString,
										 faultBlock);
		
		if (OPENSOAP_FAILED(ret)) {
			OpenSOAPEnvelopeRelease(*soapEnv);
			*soapEnv = NULL;
		}
	}

	return ret;
}


