/*-----------------------------------------------------------------------------
 * $RCSfile: XMLElm.c,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if (!defined(_WIN32)) || defined(_DEBUG)
static  const   char    CVS_ID[]        =       "$Id: XMLElm.c,v 1.64 2002/10/17 08:40:27 bandou Exp $";
#endif  /* _DEBUG */

#include <string.h>

#include <OpenSOAP/Serializer.h>

#include "XMLElm.h"
#include "XMLAttr.h"
#include "XMLNamespace.h"

/*
=begin
= OpenSOAP XML Element class
=end
 */
extern
int
OpenSOAPXMLElmReleaseMembers(/* [in, out] */ OpenSOAPXMLElmPtr elm) {
    int ret = OpenSOAPXMLNodeReleaseMembers((OpenSOAPXMLNodePtr)elm);

    if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPXMLNodeReleaseList((OpenSOAPXMLNodePtr)(elm->
															  xmlAttrList));
        if (OPENSOAP_SUCCEEDED(ret) && elm->definedXMLNsList) {
            ret = OpenSOAPXMLNamespaceReleaseList(elm->definedXMLNsList);
        }
    }

    return ret;
}

extern
int
OpenSOAPXMLElmRelease(/* [in, out] */ OpenSOAPXMLElmPtr elm) {
    int ret = OpenSOAPObjectRelease((OpenSOAPObjectPtr)elm);

    return ret;
}

static
int
OpenSOAPXMLElmFree(/* [in, out] */ OpenSOAPObjectPtr obj) {
    int ret = OpenSOAPXMLElmReleaseMembers((OpenSOAPXMLElmPtr)obj);

    if (OPENSOAP_SUCCEEDED(ret)) {
        free(obj);
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmInitialize(elm)
    Initialize XML Element Struct.

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OpenSOAPXMLElmInitialize(/* [out] */ OpenSOAPXMLElmPtr elm,
						 /* [in]  */ OpenSOAPObjectFreeFunc free_func) {

    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (elm) {
		elm->xmlAttrList = NULL;
		elm->definedXMLNsList = NULL;
        
        ret = OpenSOAPXMLNodeInitialize((OpenSOAPXMLNodePtr)elm,
                                        free_func ? free_func
                                        : OpenSOAPXMLElmFree);
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmCreate(attr)
    OpenSOAP XML Element instance create

    :Parameters
      :OpenSOAPXMLElmPtr * [out] ((|elm|))
        XML Element pointer
    :Return Values
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmCreate(/* [out] */ OpenSOAPXMLElmPtr *elm) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if(elm){
		*elm = malloc(sizeof(OpenSOAPXMLElm));

		if(*elm){
			ret = OpenSOAPXMLElmInitialize( *elm, NULL );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
		}
		else{
			ret = OPENSOAP_MEM_BADALLOC;
		}
	}
	else{
		ret = OPENSOAP_PARAMETER_BADVALUE;
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetNamespaceString(elm, nsUri, nsPrefix)
    Set Namespace of XML Element(OpenSOAPString)

    :Parameters
      :[out] OpenSOAPXMLElmPtr ((|elm|))
        XML Element.
      :[in]  OpenSOAPStringPtr ((|nsUri|))
        Namespace URI
      :[in]  OpenSOAPStringPtr ((|nsPrefix|))
        Namespace Prefix

    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
/* extern */
static
int
/* OPENSOAP_API */
OpenSOAPXMLElmSetNamespaceString(/* [out] */ OpenSOAPXMLElmPtr elm,
								 /* [in]  */ OpenSOAPStringPtr nsUri,
								 /* [in]  */ OpenSOAPStringPtr nsPrefix) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetNamespaceMB(elm, ns_uri, ns_prefix)
    Set Namespace of XML Element(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const char * [in] ((|ns_uri|))
        Namespace URI
      :const char * [in] ((|ns_prefix|))
        Namespace Prefix
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetNamespaceMB(/* [in, out] */ OpenSOAPXMLElmPtr elm,
                             /* [in]      */ const char *ns_uri,
                             /* [in]      */ const char *ns_prefix) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLNamespacePtr ns = NULL;

	ret = OpenSOAPXMLElmSearchNamespaceMB( elm, ns_uri, ns_prefix, &ns );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	/* there is a namespace defined by ns_uri and ns_prefix*/
	if( ns != NULL ){
		elm->super.thisNamespace = ns;
	}
	/* If there is no ns_uri and ns_prefix, make a new XMLNamespace,
	   input the new XMLNamespace to the definedXMLNsList,
	   input itself XMLElmPtr to definedElment */
	else{
		ret = OpenSOAPXMLNamespaceCreateMB(	ns_uri,
											ns_prefix,
											&(elm->super.thisNamespace));
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		if( !(elm->definedXMLNsList)){
			elm->definedXMLNsList = (OpenSOAPXMLNamespacePtr)(elm->super.thisNamespace);
		}
		else{
			ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)(elm->definedXMLNsList), 
											 (OpenSOAPXMLNodePtr)(elm->super.thisNamespace));
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
		}

		elm->super.thisNamespace->definedElement = elm;
	}

	ret = OPENSOAP_NO_ERROR;
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetNamespaceWC(elm, ns_uri, ns_prefix)
    Set Namespace of XML Element(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const wchar_t * [in] ((|ns_uri|))
        Namespace URI
      :const wchar_t * [in] ((|ns_prefix|))
        Namespace Prefix

    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetNamespaceWC(/* [in, out] */ OpenSOAPXMLElmPtr elm,
                             /* [in]      */ const wchar_t *ns_uri,
                             /* [in]      */ const wchar_t *ns_prefix) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSearchNamespaceString(elm, nsUri, nsPrefix, ns)
    Search Namespace of XML Element(OpenSOAPString).

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        XML Element.
      :OpenSOAPStringPtr [in] ((|nsUri|))
        Namespace URI
      :OpenSOAPStringPtr [in] ((|nsPrefix|))
        Namespace Prefix
      :OpenSOAPXMLNamespacePtr * [out] ((|ns|))
        Namespace
    :Return Value
      :int
	    Return Code Is ErrorCode
	:Note
	  This Function Is Static Function
=end
 */
/* extern */
static
int
/* OPENSOAP_API */
OpenSOAPXMLElmSearchNamespaceString(/* [in]  */ OpenSOAPXMLElmPtr elm,
                                    /* [in]  */ OpenSOAPStringPtr nsUri,
                                    /* [in]  */ OpenSOAPStringPtr nsPrefix,
                                    /* [out] */ OpenSOAPXMLNamespacePtr *ns) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (elm && ns && (nsUri || nsPrefix)) {
		OpenSOAPXMLElmPtr parent = elm;
		ret = OPENSOAP_NO_ERROR;
		*ns = NULL;
		for (; OPENSOAP_SUCCEEDED(ret) && parent && *ns;
			 parent
				 = (OpenSOAPXMLElmPtr)((OpenSOAPXMLNodePtr)parent)->parent) {
			OpenSOAPXMLNamespacePtr nextNs = parent->definedXMLNsList;
			for (; nextNs; nextNs = nextNs->next) {
				int cmpResultUri = 0;
				int cmpResultPrefix = 0;
				
				if (nsUri) {
					ret = OpenSOAPStringCompare(nextNs->uri,
												nsUri,
												&cmpResultUri);
					if (OPENSOAP_FAILED(ret)) {
						break;
					}
				}
				if (nsPrefix) {
					ret = OpenSOAPStringCompare(nextNs->prefix,
												nsPrefix,
												&cmpResultPrefix);
					if (OPENSOAP_FAILED(ret)) {
						break;
					}
				}
				/* */
				if (cmpResultPrefix == 0 && cmpResultUri != 0) {
					ret = OPENSOAP_PARAMETER_BADVALUE;
					break;
				}
				/* */
				if (cmpResultUri == 0 && cmpResultPrefix == 0) {
					*ns = nextNs;
					break;
				}
			}
		}
	}
	
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSearchNamespaceMB(elm, ns_uri, ns_prefix, ns)
    Search Namespace of XML Element(MB).

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        XML Element.
      :const char * [in] ((|ns_uri|))
        Namespace URI
      :const char * [in] ((|ns_prefix|))
        Namespace Prefix
      :OpenSOAPXMLNamespacePtr * [out] ((|ns|))
        Namespace
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSearchNamespaceMB(/* [in]  */ OpenSOAPXMLElmPtr elm,
                                /* [in]  */ const char *ns_uri,
                                /* [in]  */ const char *ns_prefix,
                                /* [out] */ OpenSOAPXMLNamespacePtr *ns) {
#if 1
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	int compare_result_uri;
	int compare_result_prefix;

	OpenSOAPXMLElmPtr next_parent = NULL;
	OpenSOAPXMLNamespacePtr next_tmp = NULL;

	if( !elm || !ns ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	/* check if there is the ns_uri and ns_prefix in itself definedXMLNsList */

	if( elm->definedXMLNsList != NULL ){
		next_tmp = elm->definedXMLNsList;
		while( next_tmp != NULL ){
			if( next_tmp->uri != NULL && ns_uri != NULL ) {
				ret = OpenSOAPStringCompareMB(next_tmp->uri, ns_uri, &compare_result_uri);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				if( next_tmp->uri == NULL && ns_uri == NULL ){
					compare_result_uri = 0;
				}
				else{
					compare_result_uri = 1;
				}
			}

			if( next_tmp->prefix != NULL && ns_prefix != NULL ) {
				ret = OpenSOAPStringCompareMB(next_tmp->prefix, ns_prefix, &compare_result_prefix);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				if( next_tmp->prefix == NULL && ns_prefix == NULL ){
					compare_result_prefix = 0;
				}
				else{
					compare_result_prefix = 1;
				}
			}

			if( compare_result_prefix == 0 ){
				if( compare_result_uri != 0 ){
					ret = OPENSOAP_PARAMETER_BADVALUE;
					return ret;
				}
				else{
					*ns = next_tmp;
					ret = OPENSOAP_NO_ERROR;
					return ret;
				}
			}
			else{
				next_tmp = next_tmp->next;
			}
		}
	}


	/* If there is no ns_uri and ns_prefix in itself definedXMLNsList,
	   check if there is the ns_uri and ns_prefix in its parent's definedXMLNsList */
	next_parent = (OpenSOAPXMLElmPtr)(((OpenSOAPXMLNodePtr)(elm))->parent);

	while( next_parent != NULL ){
		next_tmp = next_parent->definedXMLNsList;

		while( next_tmp != NULL ){
			if( next_tmp->uri != NULL && ns_uri != NULL ) {
				ret = OpenSOAPStringCompareMB( next_tmp->uri,
											   ns_uri,
											   &compare_result_uri );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				if( next_tmp->uri == NULL && ns_uri == NULL ){
					compare_result_uri = 0;
				}
				else{
					compare_result_uri = 0;
				}
			}

			if( compare_result_uri == 0 ){
				if( next_tmp->prefix != NULL && ns_prefix != NULL ) {
					ret = OpenSOAPStringCompareMB( next_tmp->prefix,
												   ns_prefix,
												   &compare_result_prefix );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
				}
				else{
					if( next_tmp->prefix == NULL && ns_prefix == NULL ){
						compare_result_prefix = 0;
					}
					else{
						compare_result_prefix = 1;
					}
				}

				if( compare_result_prefix == 0 ){
					*ns = next_tmp;
					ret = OPENSOAP_NO_ERROR;
					return ret;
				}
			}
			next_tmp = next_tmp->next;
		}

		next_parent = (OpenSOAPXMLElmPtr)(next_parent->super.parent);
	}

	/* there is no ns_uri and ns_prefix namespace */ 
	*ns = NULL;

    return ret;

#else
    OpenSOAPStringPtr ns_uri_str = NULL;
    int ret = OpenSOAPStringCreateWithMB(ns_uri, &ns_uri_str);
    if (OPENSOAP_SUCCEEDED(ret)) {
        OpenSOAPStringPtr ns_prefix_str = NULL;
        ret = OpenSOAPStringCreateWithMB(ns_prefix, &ns_prefix_str);
        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPXMLElmSearchNamespaceString(elm,
                                                      ns_uri_str,
                                                      ns_prefix_str,
                                                      ns);
            
            OpenSOAPStringRelease(ns_prefix_str);
        }
        OpenSOAPStringRelease(ns_uri_str);
    }
    
    return ret;
#endif
}


/*
=begin
--- function#OpenSOAPXMLElmSearchNamespaceWC(elm, ns_uri, ns_prefix, ns)
    Search Namespace of XML Element(WC).

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const wchar_t * [in] ((|ns_uri|))
        Namespace URI
      :const wchar_t * [in] ((|ns_prefix|))
        Namespace Prefix
      :OpenSOAPXMLNamespacePtr * [out] ((|ns|))
        Namespace
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSearchNamespaceWC(/* [in]  */ OpenSOAPXMLElmPtr elm,
                                /* [in]  */ const wchar_t *ns_uri,
                                /* [in]  */ const wchar_t *ns_prefix,
                                /* [out] */ OpenSOAPXMLNamespacePtr *ns) {
    OpenSOAPStringPtr ns_uri_str = NULL;
    int ret = OpenSOAPStringCreateWithWC(ns_uri, &ns_uri_str);
    if (OPENSOAP_SUCCEEDED(ret)) {
        OpenSOAPStringPtr ns_prefix_str = NULL;
        ret = OpenSOAPStringCreateWithWC(ns_prefix, &ns_prefix_str);
        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPXMLElmSearchNamespaceString(elm,
                                                      ns_uri_str,
                                                      ns_prefix_str,
                                                      ns);
            
            OpenSOAPStringRelease(ns_prefix_str);
        }
        OpenSOAPStringRelease(ns_uri_str);
    }
    
    return ret;
}


/*
=begin
--- function#OpenSOAPXMLElmGetNamespace(elm, ns)
    Get Namespace Pointer

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element
      :OpenSOAPXMLNamespacePtr * [out] ((|ns|))
        Namespace 
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetNamespace(/* [in]  */ OpenSOAPXMLElmPtr elm,
                           /* [out] */ OpenSOAPXMLNamespacePtr *ns) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !elm || !ns ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
	}
	else{
		*ns = elm->super.thisNamespace;
		ret = OPENSOAP_NO_ERROR;
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmDefineNamespaceMB(elm, ns_uri, ns_prefix, ns)
    Define XML namespace.(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const char * [in] ((|ns_uri|))
        Namespace URI
      :const char * [in] ((|ns_prefix|))
        Namespace Prefix
      :OpenSOAPXMLNamespacePtr * [out] ((|ns|))
        OpenSOAP XML Namespace.
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmDefineNamespaceMB(/* [out] */ OpenSOAPXMLElmPtr elm,
                                /* [in]  */ const char *ns_uri,
                                /* [in]  */ const char *ns_prefix,
                                /* [out] */ OpenSOAPXMLNamespacePtr *ns) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLNamespacePtr ns_next = NULL;

	int compare_result_uri;
	int compare_result_prefix;

	if( !elm || !ns ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( elm->definedXMLNsList != NULL ){
		ns_next = elm->definedXMLNsList;
		while( ns_next != NULL ){
			if( ns_next->uri != NULL && ns_uri != NULL ){
				ret = OpenSOAPStringCompareMB(ns_next->uri, ns_uri, &compare_result_uri);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				if( ns_next->uri == NULL && ns_uri == NULL ){
					compare_result_uri = 0;
				}
				else{
					compare_result_uri = 1;
				}
			}

			if( ns_next->prefix != NULL && ns_prefix != NULL ){
				ret = OpenSOAPStringCompareMB(ns_next->prefix, ns_prefix, &compare_result_prefix);
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				if( ns_next->prefix == NULL && ns_prefix == NULL ){
					compare_result_prefix = 0;
				}
				else{
					compare_result_prefix = 1;
				}
			}

			if( compare_result_prefix == 0 ){
				if( compare_result_uri != 0 ){
					ret = OPENSOAP_PARAMETER_BADVALUE;
					return ret;
				}
				else{
					*ns = ns_next;
					ret = OPENSOAP_NO_ERROR;
					return ret;
				}
			}
			else{
				ns_next = ns_next->next;
			}
		}
	}

    ret = OpenSOAPXMLNamespaceCreateMB(	ns_uri,
										ns_prefix,
										ns);
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	if( !(elm->definedXMLNsList)){
		elm->definedXMLNsList = *ns;
	}
	else{
		ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)(elm->definedXMLNsList), 
										 (OpenSOAPXMLNodePtr)(*ns));
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
	}

	(*ns)->definedElement = elm;

	ret = OPENSOAP_NO_ERROR;

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmDefineNamespaceWC(elm, ns_uri, ns_prefix, ns)
    Define XML namespace.(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const wchar_t * [in] ((|ns_uri|))
        Namespace URI
      :const wchar_t * [in] ((|ns_prefix|))
        Namespace Prefix
      :OpenSOAPXMLNamespacePtr * [out] ((|ns|))
        OpenSOAP XML Namespace.
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmDefineNamespaceWC(/* [out] */ OpenSOAPXMLElmPtr elm,
                                /* [in]  */ const wchar_t *ns_uri,
                                /* [in]  */ const wchar_t *ns_prefix,
                                /* [out] */ OpenSOAPXMLNamespacePtr *ns) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddAttributeString(elm, attr_name, attr_type, attr_value, attr)
    Add Attribute(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        SOAP Element.
      :OpenSOAPStringPtr [in] ((|attr_name|))
        Attribute Name
      :OpenSOAPStringPtr [in] ((|attr_type|))
        Attribute Type
	  :void * [in] ((|attr_value|))
        Attribute Value
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        OpenSOAP XML Attribute
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmAddAttributeString(/* [out] */ OpenSOAPXMLElmPtr elm,
								 /* [in]  */ OpenSOAPStringPtr attr_name,
								 /* [in]  */ OpenSOAPStringPtr attr_type,
								 /* [in]  */ void *attr_value,
								 /* [out] */ OpenSOAPXMLAttrPtr *attr) {

    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLAttrPtr next_attr;

	if( elm ){
		ret = OpenSOAPXMLAttrCreateString( attr_name, attr );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
		ret = OpenSOAPXMLAttrSetValueString( *attr, attr_type, attr_value );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		(*attr)->definedElement = elm;

		next_attr = elm->xmlAttrList;
		if( !next_attr ){
			elm->xmlAttrList = *attr;
		}
		else{
			ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)next_attr,
											 (OpenSOAPXMLNodePtr)(*attr) );
		}
	}
	else{
		ret = OPENSOAP_PARAMETER_BADVALUE;
	}
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddAttributeMB(elm, attr_name, attr_type, attr_value, attr)
    Add Attribute(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        SOAP Element.
      :const char * [in] ((|attr_name|))
        Attribute Name
      :const char * [in] ((|attr_type|))
        Attribute Type
      :void * [in] ((|attr_value|))
        Attribute Value
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        OpenSOAP XML Attribute
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmAddAttributeMB(/* [out] */ OpenSOAPXMLElmPtr elm,
                             /* [in]  */ const char *attr_name,
                             /* [in]  */ const char *attr_type,
                             /* [in]  */ void *attr_value,
                             /* [out] */ OpenSOAPXMLAttrPtr *attr) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLAttrPtr next_attr;

	if( elm ){
		ret = OpenSOAPXMLAttrCreateMB( attr_name, attr );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		ret = OpenSOAPXMLAttrSetValueMB( *attr, attr_type, attr_value );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		(*attr)->definedElement = elm;

		next_attr = elm->xmlAttrList;
		if( !next_attr ){
			elm->xmlAttrList = *attr;
		}
		else{
			ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)next_attr,
											 (OpenSOAPXMLNodePtr)(*attr) );
		}

	}
	else{
		ret = OPENSOAP_PARAMETER_BADVALUE;
	}

    return ret;
}


/*
=begin
--- function#OpenSOAPXMLElmAddAttributeWC(elm, attr_name, attr_type, attr_value, attr)
    Add Attribute(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        SOAP Element.
      :const wchar_t * [in] ((|attr_name|))
        Attribute Name
      :const wchar_t * [in] ((|attr_type|))
        Attribute Type
	  :void * [in] ((|attr_value|))
        Attribute Value
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        OpenSOAP XML Attribute
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmAddAttributeWC(/* [out] */ OpenSOAPXMLElmPtr elm,
                             /* [in]  */ const wchar_t *attr_name,
                             /* [in]  */ const wchar_t *attr_type,
                             /* [in]  */ void *attr_value,
                             /* [out] */ OpenSOAPXMLAttrPtr *attr) {
	OpenSOAPStringPtr attr_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(attr_name, &attr_name_str);

	if (OPENSOAP_SUCCEEDED(ret)) {
		OpenSOAPStringPtr attr_type_str = NULL;
		ret = OpenSOAPStringCreateWithWC(attr_type, &attr_type_str);
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = OpenSOAPXMLElmAddAttributeString(elm,
												   attr_name_str,
												   attr_type_str,
												   attr_value, attr);
			OpenSOAPStringRelease(attr_type_str);
		}
		OpenSOAPStringRelease(attr_name_str);
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetAttributeString(elm, attr_name, attr)
    Get Attribute Pointer of Assignment Attribute Name(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :OpenSOAPStringPtr [in] ((|attr_name|))
        Attribute Name
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        OpenSOAP XML Attribute
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetAttributeString(/* [in]  */ OpenSOAPXMLElmPtr elm,
								 /* [in]  */ OpenSOAPStringPtr attr_name,
								 /* [out] */ OpenSOAPXMLAttrPtr *attr) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;
	int compare_result;

	OpenSOAPXMLAttrPtr  next_attr;

	if( !elm || !attr ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	next_attr = elm->xmlAttrList;

	while( next_attr != NULL ){
		ret = OpenSOAPStringCompare( next_attr->super.name, attr_name, &compare_result );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		if( compare_result == 0 ){
			*attr = next_attr;
			ret = OPENSOAP_NO_ERROR;
			return ret;
		}
		next_attr = (OpenSOAPXMLAttrPtr)(next_attr->super.next);
	}

	/* Cannot Find */
	if( next_attr == NULL ){
		*attr = NULL;
		ret = OPENSOAP_NO_ERROR;
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetAttributeMB(elm, attr_name, attr)
    Get Attribute Pointer of Assignment Attribute Name(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const char * [in] ((|attr_name|))
        Attribute Name
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        OpenSOAP XML Attribute
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetAttributeMB(/* [in]  */ OpenSOAPXMLElmPtr elm,
                             /* [in]  */ const char *attr_name,
                             /* [out] */ OpenSOAPXMLAttrPtr *attr) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;
	int compare_result;

	OpenSOAPXMLAttrPtr  next_attr;

	if( !elm || !attr ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	next_attr = elm->xmlAttrList;

	while( next_attr != NULL ){
		ret = OpenSOAPStringCompareMB( next_attr->super.name, attr_name, &compare_result );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		if( compare_result == 0 ){
			*attr = next_attr;
			ret = OPENSOAP_NO_ERROR;
			return ret;
		}
		next_attr = (OpenSOAPXMLAttrPtr)(next_attr->super.next);
	}

	/* Cannot Find */
	if( next_attr == NULL ){
		*attr = NULL;
		ret = OPENSOAP_NO_ERROR;
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetAttributeWC(elm, attr_name, attr)
    Get Attribute Pointer of Assignment Attribute Name(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        XML Element.
      :const wchar_t * [in] ((|attr_name|))
        Attribute Name
      :OpenSOAPXMLAttrPtr * [out] ((|attr|))
        OpenSOAP XML Attribute
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetAttributeWC(/* [out] */ OpenSOAPXMLElmPtr elm,
                             /* [in]  */ const wchar_t *attr_name,
                             /* [out] */ OpenSOAPXMLAttrPtr *attr) {
	OpenSOAPStringPtr attr_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(attr_name, &attr_name_str);

	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPXMLElmGetAttributeString(elm, attr_name_str, attr);
		OpenSOAPStringRelease(attr_name_str);
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetNextChild(elm, cld_elm)
    Get next child XML Element.

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :OpenSOAPXMLElmPtr * [in, out] ((|cld_elm|))
        Next XML Element Pointer of Assignment XML Element
        But If *cld_elm as in NULL then Return To First XML Element Pointer 
    :Return Value value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetNextChild(/* [in]      */ OpenSOAPXMLElmPtr elm,
                           /* [in, out] */ OpenSOAPXMLElmPtr *cld_elm) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLNodePtr  child_next;

	if( !elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( !(*cld_elm) ){
		*cld_elm = (OpenSOAPXMLElmPtr)(elm->super.children);
		ret = OPENSOAP_NO_ERROR;
	}
	else{
		child_next = elm->super.children;
		while( child_next != NULL ){
			if( child_next == (OpenSOAPXMLNodePtr)(*cld_elm) ){
				*cld_elm = (OpenSOAPXMLElmPtr)((*cld_elm)->super.next);
				ret = OPENSOAP_NO_ERROR;
				return ret;
			}
			child_next = child_next->next;
		}
	}
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddChildString(elm, cld_name, cld_elm)
    Add Child XML Element(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :OpenSOAPStringPtr [in] ((|cld_name|))
        Add Child Element Name
      :OpenSOAPXMLElmPtr * [out] ((|cld_elm|))
        Add Child Element Pointer
    :Return Value value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmAddChildString(/* [in]  */ OpenSOAPXMLElmPtr elm,
							 /* [in]  */ OpenSOAPStringPtr childName,
							 /* [in, out] */ OpenSOAPXMLElmPtr *childElm) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;
	
	if (elm && childName && childElm) {
		ret = (*childElm) ? OPENSOAP_NO_ERROR
			: OpenSOAPXMLElmCreate(childElm);
		if (OPENSOAP_SUCCEEDED(ret)) {
			OpenSOAPXMLNodePtr node = (OpenSOAPXMLNodePtr)elm;
			OpenSOAPXMLNodePtr childNode = (OpenSOAPXMLNodePtr)*childElm;
			OpenSOAPStringPtr childElmName = NULL;
			ret = OpenSOAPStringDuplicate(childName, &childElmName);
			if (OPENSOAP_SUCCEEDED(ret)) {
				if (node->children) {
					ret = OpenSOAPXMLNodeInsertNext(node->children,
													childNode);
				}
				else {
					node->children = childNode;
				}
				if (OPENSOAP_SUCCEEDED(ret)) {
					/* swap */
					OpenSOAPStringPtr prevChildName = childNode->name;
					childNode->name = childElmName;
					childElmName = prevChildName;
					/* parent setting */
					childNode->parent = node;
				}
				OpenSOAPStringRelease(childElmName);
			}

			if (OPENSOAP_FAILED(ret)) {
				OpenSOAPXMLElmRelease(*childElm);
			}
		}
	}
	
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddChildMB(elm, cld_name, cld_elm)
    Add Child XML Element(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :cosnt char * [in] ((|cld_name|))
        Add Child Element Name
      :OpenSOAPXMLElmPtr * [out] ((|cld_elm|))
        Add Child Element Pointer
    :Return Value value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmAddChildMB(/* [out] */ OpenSOAPXMLElmPtr elm,
						 /* [in]  */ const char *cld_name,
						 /* [out] */ OpenSOAPXMLElmPtr *cld_elm) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	if( !elm || !cld_elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( !*cld_elm ){
		ret = OpenSOAPXMLElmCreate( cld_elm );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
	}

	ret = OpenSOAPStringCreate( &((*cld_elm)->super.name) );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

    ret = OpenSOAPStringSetStringMB((*cld_elm)->super.name, cld_name);
	if(ret == OPENSOAP_PARAMETER_BADVALUE){
		return ret;
	}

	(*cld_elm)->super.parent = (OpenSOAPXMLNodePtr)elm;

	if( !elm->super.children ){
		elm->super.children = (OpenSOAPXMLNodePtr)(*cld_elm);
	}
	else{
		ret = OpenSOAPXMLNodeInsertNext( (OpenSOAPXMLNodePtr)(elm->super.children),
										 (OpenSOAPXMLNodePtr)(*cld_elm) );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}
	}

	ret = OPENSOAP_NO_ERROR;
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddChildWC(elm, cld_name, cld_elm)
    Add Child XML Element(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :cosnt wchar_t * [in] ((|cld_name|))
        Add Child Element Name
      :OpenSOAPXMLElmPtr * [out] ((|cld_elm|))
        Add Child Element Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmAddChildWC(/* [out] */ OpenSOAPXMLElmPtr elm,
						 /* [in]  */ const wchar_t *cld_name,
						 /* [out] */ OpenSOAPXMLElmPtr *cld_elm) {
	OpenSOAPStringPtr cld_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(cld_name, &cld_name_str);

	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPXMLElmAddChildString(elm, cld_name_str, cld_elm);
		OpenSOAPStringRelease(cld_name_str);
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetChildString(elm, chldName, chldElm)
    Get Child XML Element(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :OpenSOAPStringPtr [in] ((|chldName|))
        Assignment Element Name 
      :OpenSOAPXMLElmPtr * [out] ((|chldElm|))
        XML Element Pointer of Assignment Element Name 
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetChildString(/* [in]  */ OpenSOAPXMLElmPtr elm,
							 /* [in]  */ OpenSOAPStringPtr chldName,
							 /* [out] */ OpenSOAPXMLElmPtr *chldElm) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (elm && chldName && chldElm) {
		OpenSOAPXMLNodePtr  child = ((OpenSOAPXMLNodePtr)elm)->children;
		for (ret = OPENSOAP_NO_ERROR; child; child = child->next) {
			int cmpResult = 0;
			ret = OpenSOAPStringCompare(child->name,
										chldName,
										&cmpResult);
			if (OPENSOAP_FAILED(ret)
				|| cmpResult == 0) {
				break;
			}
		}
		if (OPENSOAP_SUCCEEDED(ret)) {
			*chldElm = (OpenSOAPXMLElmPtr)child;
		}
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetChildMB(elm, cld_name, cld_elm)
    Get Child XML Element(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :cosnt char * [in] ((|cld_name|))
        Assignment Element Name 
      :OpenSOAPXMLElmPtr * [out] ((|cld_elm|))
        XML Element Pointer of Assignment Element Name 
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetChildMB(/* [in]  */ OpenSOAPXMLElmPtr elm,
						 /* [in]  */ const char *cld_name,
						 /* [out] */ OpenSOAPXMLElmPtr *cld_elm) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;
	int compare_result;

	OpenSOAPXMLNodePtr  child;

	if( !elm || !cld_elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	child = elm->super.children;

	while( child != NULL ){
		ret = OpenSOAPStringCompareMB( child->name, cld_name, &compare_result );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		if( compare_result == 0 ){
			*cld_elm = ( OpenSOAPXMLElmPtr ) child;
			ret = OPENSOAP_NO_ERROR;
			return ret;
		}
		child = child->next;
	}

	// Cannot Find
	if( child == NULL ){
		*cld_elm = NULL;
		ret = OPENSOAP_NO_ERROR;
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetChildWC(elm, cld_name, cld_elm)
    Get Child XML Element(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :cosnt wchar_t * [in] ((|cld_name|))
        Assignment Element Name 
      :OpenSOAPXMLElmPtr * [out] ((|cld_elm|))
        XML Element Pointer of Assignment Element Name 
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetChildWC(/* [in]  */ OpenSOAPXMLElmPtr elm,
						 /* [in]  */ const wchar_t *cld_name,
						 /* [out] */ OpenSOAPXMLElmPtr *cld_elm) {
	OpenSOAPStringPtr cld_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(cld_name, &cld_name_str);

	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPXMLElmGetChildString(elm, cld_name_str, cld_elm);
		OpenSOAPStringRelease(cld_name_str);
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetValueString(elm, typeName, value)
    Get Value of XML Element(OpenSOAPString)

    :Parameters
      :[in]  OpenSOAPXMLElmPtr ((|elm|))
        OpenSOAP XML Element.
      :[in]  OpenSOAPString ((|typeName|))
        Type Name
      :[out] void * ((|value|))
        Setting Buffer Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetValueString(/* [in]  */ OpenSOAPXMLElmPtr elm,
							 /* [in]  */ OpenSOAPStringPtr typeName,
							 /* [out] */ void *value) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;
	
	if (elm && typeName && value) {
		OpenSOAPDeserializerFunc deserializer = NULL;
		ret = OpenSOAPGetDeserializer(typeName, &deserializer);
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = deserializer(((OpenSOAPXMLNodePtr)elm)->value, value);
		}
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetValueMB(elm, typeName, value)
    Get Value of XML Element(MB)

    :Parameters
      :[in]  OpenSOAPXMLElmPtr ((|elm|))
        OpenSOAP XML Element.
      :[in]  const char * ((|type_name|))
        Type Name
      :[out] void * ((|value|))
        Setting Buffer Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetValueMB(/* [in]  */ OpenSOAPXMLElmPtr elm,
                         /* [in]  */ const char *typeName,
                         /* [out] */ void *value) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (elm) {
		OpenSOAPDeserializerFunc deserializer = NULL;
		/* get deserializer */
		ret = OpenSOAPGetDeserializerMB(typeName, &deserializer);
		if (OPENSOAP_SUCCEEDED(ret)) {
			/* deserialize */
			ret = deserializer(((OpenSOAPXMLNodePtr)elm)->value,
							   value);
		}
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetValueWC(elm, type_name, value)
    Get Value of XML Element(WC)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :const wchar_t * [in] ((|type_name|))
        Type Name
      :void * [out] ((|value|))
        Setting Buffer Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetValueWC(/* [in]  */ OpenSOAPXMLElmPtr elm,
                         /* [in]  */ const wchar_t *type_name,
                         /* [out] */ void *value) {
	OpenSOAPStringPtr type_name_str = NULL;
	int ret = OpenSOAPStringCreateWithWC(type_name, &type_name_str);

	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPXMLElmGetValueString(elm, type_name_str, value);
		OpenSOAPStringRelease(type_name_str);
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetValueString(elm, typeName, value)
    Setting Value of XML Element(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :OpenSOAPStringPtr [in] ((|typeName|))
        Type Name
      :void * [in] ((|value|))
        Setting Buffer Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetValueString(/* [out] */ OpenSOAPXMLElmPtr elm,
							 /* [in]  */ OpenSOAPStringPtr typeName,
							 /* [in]  */ void *value) {
	int ret = OpenSOAPXMLNodeSetValueString((OpenSOAPXMLNodePtr)elm,
											typeName,
											value);

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetValueMB(elm, typeName, value)
    Setting Value of XML Element(MB)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :const char * [in] ((|typeName|))
        Type Name
      :void * [in] ((|value|))
        Setting Buffer Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetValueMB(/* [in] */ OpenSOAPXMLElmPtr elm,
                         /* [in] */ const char *typeName,
                         /* [in] */ void *value) {
	int ret = OpenSOAPXMLNodeSetValueMB((OpenSOAPXMLNodePtr)elm,
										typeName,
										value);

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetValueWC(elm, typeName, value)
    Setting Value of XML Element(WC)

    :Parameters
      :[out] OpenSOAPXMLElmPtr ((|elm|))
        OpenSOAP XML Element.
      :[in]  const wchar_t * ((|typeName|))
        Type Name
      :[in]  void * ((|value|))
        Setting Buffer Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetValueWC(/* [out] */ OpenSOAPXMLElmPtr elm,
                         /* [in]  */ const wchar_t *typeName,
                         /* [in]  */ void *value) {
	int ret = OpenSOAPXMLNodeSetValueWC((OpenSOAPXMLNodePtr)elm,
										typeName,
										value);

	return ret;
}

/*
=begin
--- function#OpenSOAPLibxmlDocDumpByteArray(doc, chEnc, bAry)
    Libxml's doc dump to OpenSOAPByteArray.

    :Parameters
      :[in]  xmlDocPtr ((|doc|))
        libxml document pointer.
      :[in]  const char  *((|chEnc|))
        character encoding. i.e. "EUC-JP", "SHIFT-JIS", "UTF-8"
      :[out] OpenSOAPByteArrayPtr ((|bAry|))
        Result Buffer        
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
static
int
OpenSOAPLibxmlDocDumpByteArray(/* [in]  */ xmlDocPtr	doc,
							   /* [in]  */ const char *chEnc,
							   /* [out] */ OpenSOAPByteArrayPtr bAry) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (doc) {
		const char XML_PROC_BEGIN[] = "<?xml";
		const char XML_PROC_END[] = "?>";
		xmlChar *xmlText = NULL;
		xmlChar *xmlTextEnd = NULL;
		size_t xmlTextLen = 0;
		xmlChar *xmlBodyBegin = NULL;
		
		/* change xmlDoc to text */
#if 1
		xmlKeepBlanksDefault(0);
		xmlDocDumpFormatMemory(doc,
							   &xmlText,
							   &xmlTextLen,
							   1);
#else		
		xmlDocDumpMemory(doc, &xmlText, &xmlTextLen);
#endif
		if (!xmlText) {
			ret = OPENSOAP_MEM_BADALLOC; /* ? */
			return ret;
		}
		ret = OPENSOAP_NO_ERROR;
		/* */
		xmlTextEnd = xmlText + xmlTextLen;
		/* */
		xmlBodyBegin = xmlText;

		/* delete the processing line <?xml ... ?> */
		if (strncmp(xmlBodyBegin,
					XML_PROC_BEGIN, sizeof(XML_PROC_BEGIN) - 1) == 0) {
			/* have xml processing line */
			char *xmlProcEnd = strstr(xmlBodyBegin, XML_PROC_END);
			if (!xmlProcEnd) {
				ret = OPENSOAP_XML_BADMAKEDOCUMENT;
			}
			else {
				const char OMITTED_CHARS[] = "\r\n";
				xmlBodyBegin = xmlProcEnd + sizeof(XML_PROC_END) - 1;

				/* remove CR/LF */
				for (; xmlBodyBegin != xmlTextEnd
						 && strchr(OMITTED_CHARS, *xmlBodyBegin);
					 ++xmlBodyBegin) {
				}
			}
		}

		if (OPENSOAP_SUCCEEDED(ret)) {
			OpenSOAPByteArrayPtr utf8XmlText = NULL;
			/* create ByteArray */
			ret = OpenSOAPByteArrayCreate(&utf8XmlText);
			if (OPENSOAP_SUCCEEDED(ret)) {
				/* set UTF-8 encode character data */
				ret = OpenSOAPStringConvertXMLCharRefToUTF8(xmlBodyBegin,
															utf8XmlText);
				if (OPENSOAP_SUCCEEDED(ret)) {
					/* convert UTF-8 to chEnc character encoding */
					ret = OpenSOAPStringConvertCharEncoding("UTF-8",
															utf8XmlText,
															chEnc,
															bAry);
				}
			}
			/* release temporary byte array */
			OpenSOAPByteArrayRelease(utf8XmlText);
		}
		/* */
		xmlFree(xmlText);
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLAttrListAddToLibxml(attrList, doc, elm)
    Add attribute to the element node in libxml structure from opensoap structure 
    
    :Parameters
      :[in]  OpenSOAPXMLAttrPtr ((|attrList|))
        OpenSOAP XML Attribute list
      :[out] xmlDocPtr ((|doc|))
        Libxml Document Pointer
      :[out] xmlNodePtr ((|elm|))
        Libxml Node Pointer
    :Return value
      :type int
        error code.
=end
 */
static
int
OpenSOAPXMLAttrListAddToLibxml(/* [in]  */ OpenSOAPXMLAttrPtr attrList,
							   /* [out] */ xmlDocPtr doc,
							   /* [out] */ xmlNodePtr elm) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (elm && doc) {
		char *utf8AttrName = NULL;
		char *utf8AttrValue = NULL;
		char *utf8NsPrefix = NULL;
		char *utf8NsURI = NULL;
		OpenSOAPXMLNodePtr attr = (OpenSOAPXMLNodePtr)attrList;
		for (ret = OPENSOAP_NO_ERROR; attr; attr = attr->next) {
			OpenSOAPXMLNamespacePtr attrNs = attr->thisNamespace;
			xmlNsPtr libxmlAttrNs = NULL;
			xmlAttrPtr libxmlAttr = NULL;
			/* get name */
			ret = OpenSOAPStringGetStringUTF8WithAllocator(attr->name,
														   NULL,
														   NULL,
														   &utf8AttrName);
			if (OPENSOAP_FAILED(ret)) {
				break;
			}
			/* get value */
			ret = OpenSOAPStringGetStringUTF8WithAllocator(attr->value,
														   NULL,
														   NULL,
														   &utf8AttrValue);
			if (OPENSOAP_FAILED(ret)) {
				break;
			}
			
			if (attrNs) {
				/* get uri */
				ret = OpenSOAPStringGetStringUTF8WithAllocator(attrNs->uri,
															   NULL,
															   NULL,
															   &utf8NsURI);
				if (OPENSOAP_FAILED(ret)) {
					break;
				}

				/* get prefix */
				ret = OpenSOAPStringGetStringUTF8WithAllocator(attrNs->prefix,
															   NULL,
															   NULL,
															   &utf8NsPrefix);
				if (OPENSOAP_FAILED(ret)) {
					break;
				}

				/* search Namespace */
				libxmlAttrNs = xmlSearchNs(doc, elm, utf8NsPrefix);
				if (!libxmlAttrNs
					|| (!libxmlAttrNs->href && !utf8NsURI)
					|| !(libxmlAttrNs->href && utf8NsURI
						&& (strcmp(libxmlAttrNs->href, utf8NsURI) == 0))) {
					libxmlAttrNs = xmlNewNs(elm,
											utf8NsURI,
											utf8NsPrefix);
				}
			}

			libxmlAttr = libxmlAttrNs
				? xmlNewNsProp(elm, libxmlAttrNs, utf8AttrName, utf8AttrValue)
				: xmlSetProp(elm, utf8AttrName, utf8AttrValue);

			free(utf8NsURI);
			utf8NsURI = NULL;
			free(utf8NsPrefix);
			utf8NsPrefix = NULL;
			free(utf8AttrValue);
			utf8AttrValue = NULL;
			free(utf8AttrName);
			utf8AttrName = NULL;
		}
		free(utf8NsURI);
		free(utf8NsPrefix);
		free(utf8AttrValue);
		free(utf8AttrName);
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddAttrInEncodingString(doc, elmChild, childNode)
    Add attribute to the element node in libxml structure from opensoap structure 
    
    :Parameters
      :xmlDocPtr [in ] ((|doc|))
        Libxml Document Pointer
      :OpenSOAPXMLElmPtr [in] ((|elmChild|))
        OpenSOAP Element Pointer
      :xmlNodePtr [in, out] ((|childNode|))
        Libxml Node Pointer
    :Return value
      :type int
        error code.
      
=end
*/
static
int
OpenSOAPXMLElmAddAttrInEncodingString(xmlDocPtr doc,
									  OpenSOAPXMLElmPtr elmChild,
									  xmlNodePtr childNode) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (elmChild) {
		ret = OpenSOAPXMLAttrListAddToLibxml(elmChild->xmlAttrList,
											 doc,
											 childNode);
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetCharEncodingString(elm, chEnc, bAry)
    XML Element character encoding output

    :Parameters
      :[in]  OpenSOAPXMLElmPtr ((|elm|))
        OpenSOAP XML Element.
      :[in]  const char  *((|chEnc|))
        character encoding. i.e. "EUC-JP", "SHIFT-JIS", "UTF-8"
      :[out] OpenSOAPByteArrayPtr ((|bAry|))
        Result Buffer        
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetCharEncodingString(/* [in]  */ OpenSOAPXMLElmPtr elm,
                                    /* [in]  */ const char *chEnc,
                                    /* [out] */ OpenSOAPByteArrayPtr bAry) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (elm) {
		xmlDocPtr	doc = NULL;
		OpenSOAPXMLNodePtr node = (OpenSOAPXMLNodePtr)elm;
		char *utf8NodeName = NULL;
		char *utf8NodeValue = NULL;

		/* make a new xmlDoc */ 
		doc = xmlNewDoc("1.0");

		ret = OpenSOAPStringGetStringUTF8WithAllocator(node->name,
													   NULL,
													   NULL,
													   &utf8NodeName);
		if (OPENSOAP_SUCCEEDED(ret)) {
			if (node->value) {
				ret = OpenSOAPStringGetStringUTF8WithAllocator(node->value,
															   NULL,
															   NULL,
															   &utf8NodeValue);
			}
			if (OPENSOAP_SUCCEEDED(ret)) {
				xmlNodePtr	elmNode = xmlNewDocNode(doc, 
													NULL, 
													utf8NodeName,
													utf8NodeValue);
				doc->children = elmNode;
				
				/* add root node's namespace */
				ret = OpenSOAPXMLElmAddNsInEncodingString(doc,
														  elm,
														  elmNode);
				if (OPENSOAP_SUCCEEDED(ret)) {
					/* add root node's attributes */
					ret = OpenSOAPXMLElmAddAttrInEncodingString(doc,
																elm,
																elmNode);
					if (OPENSOAP_SUCCEEDED(ret)) {
						/* add children node */
						ret =  OpenSOAPXMLElmAddNodeWithLibxml(doc,
															   elmNode,
															   elm);
						if (OPENSOAP_SUCCEEDED(ret)) {
							/* libxml document to byte array */
							ret = OpenSOAPLibxmlDocDumpByteArray(doc,
																 chEnc,
																 bAry);
						}
					}
				}
			}
		}

		/* free utf8 strings */
		free(utf8NodeValue);
		free(utf8NodeName);
		/* free libxml document */
		xmlFreeDoc(doc);
	}

    return ret;
}


/*
=begin
--- function#OpenSOAPXMLElmAddNodeWithLibxml(doc, node, elm)
    create libxml structure from opensoap structure 
    
    :Parameters
      :xmlDocPtr [in ] ((|doc|))
        Libxml Document Pointer
      :xmlNodePtr [in, out] ((|node|))
        Libxml Node Pointer
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP Element Pointer
    :Return value
      :type int
        error code.
      
=end
*/
int
OpenSOAPXMLElmAddNodeWithLibxml(/* [out] */ xmlDocPtr doc,
								/* [out] */ xmlNodePtr node,
								/* [in]  */	OpenSOAPXMLElmPtr elm) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;
	/* parameter check */
	if (elm && node && doc) {
		char *utf8NodeName = NULL;
		char *utf8NodeValue = NULL;
		OpenSOAPXMLNodePtr	nodeItr = ((OpenSOAPXMLNodePtr)elm)->children;
		for (ret = OPENSOAP_NO_ERROR; nodeItr; nodeItr = nodeItr->next) {
			OpenSOAPXMLElmPtr childElm = (OpenSOAPXMLElmPtr)nodeItr;
			xmlNodePtr	childNode = NULL;
			if (nodeItr->value) {
				ret = OpenSOAPStringGetStringUTF8WithAllocator(nodeItr->value,
															   NULL,
															   NULL,
															   &utf8NodeValue);
				if (OPENSOAP_FAILED(ret)) {
					break;
				}
			}
			/* */
			ret = OpenSOAPStringGetStringUTF8WithAllocator(nodeItr->name,
														   NULL,
														   NULL,
														   &utf8NodeName);
			if (OPENSOAP_FAILED(ret)) {
				break;
			}
			/* */
			childNode = xmlNewChild(node,
									NULL,
									utf8NodeName,
									utf8NodeValue);
			
			/* add element's namespace */
			ret = OpenSOAPXMLElmAddNsInEncodingString(doc,
													  childElm,
													  childNode);
			if (OPENSOAP_FAILED(ret)) {
				break;
			}

			/* add attributes */
			ret = OpenSOAPXMLElmAddAttrInEncodingString(doc,
														childElm,
														childNode);
			if (OPENSOAP_FAILED(ret)) {
				break;
			}

			/* add node */
			ret = OpenSOAPXMLElmAddNodeWithLibxml(doc, childNode, childElm);
			if (OPENSOAP_FAILED(ret)) {
				break;
			}

			free(utf8NodeName);
			utf8NodeName = NULL;
			free(utf8NodeValue);
			utf8NodeValue = NULL;
		}

		free(utf8NodeName);
		free(utf8NodeValue);
	}
	
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmIsParentDefinedNs(elm, ns, result)
    check parent element's defined namespace
    
    :Parameters
      :[in]  OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP Element Pointer
      :[in]  OpenSOAPXMLNamespacePtr ((|ns|))
        OpenSOAP XML Namespace Pointer
      :[out] int * ((|result|))
        result return buffer. If *result is non zero,
		then ns is defined at parents.
    :Return value
      :type int
        error code.
      
=end
*/
static
int
OpenSOAPXMLElmIsParentDefinedNs(/* [in]  */ OpenSOAPXMLElmPtr elm,
								/* [in]  */ OpenSOAPXMLNamespacePtr ns,
								/* [out] */ int *result) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (elm && ns && result) {
		OpenSOAPXMLNodePtr parent = (OpenSOAPXMLNodePtr)elm;

		ret = OPENSOAP_NO_ERROR;
		*result = 0;

		for (parent = parent->parent;
			parent; parent = parent->parent) {
			OpenSOAPXMLNamespacePtr nsItr
				= ((OpenSOAPXMLElmPtr)parent)->definedXMLNsList;
			for (; nsItr && nsItr != ns; nsItr = nsItr->next) {
			}
			if (nsItr) {
				*result = 1;
				break;
			}
		}
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLNamespaceGetUTF8Properties(ns, memAllocator, utf8NsURI, utf8NsPrefix)
    get XML Namespace properties UTF-8 encoding
    
    :Parameters
      :[in]  OpenSOAPXMLNamespacePtr ((|ns|))
        OpenSOAP XML Namespace Pointer
      :[in]  char *(* ((|memAllocator|)) )(size_t)
        memory allocator
      :[out] char ** ((|utf8NsURI|))
        URI stored buffer.
      :[out] char ** ((|utf8NsPrefix|))
        prefix stored buffer.
    :Return value
      :type int
        error code.
      
=end
*/
static
int
OpenSOAPXMLNamespaceGetUTF8Properties(/* [in]  */ OpenSOAPXMLNamespacePtr ns,
									  /* [in]  */ char *(*memAllocator)(size_t),
									  /* [out] */ char **utf8NsURI,
									  /* [out] */ char **utf8NsPrefix) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;
	if (ns && utf8NsURI && utf8NsPrefix) {
		ret = (ns->uri)
			? OpenSOAPStringGetStringUTF8WithAllocator(ns->uri,
													   memAllocator,
													   NULL,
													   utf8NsURI)
			: OPENSOAP_NO_ERROR;
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = (ns->prefix)
				? OpenSOAPStringGetStringUTF8WithAllocator(ns->prefix,
														   memAllocator,
														   NULL,
														   utf8NsPrefix)
				: OPENSOAP_NO_ERROR;
		}
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetNsToLibxmlElm(childElm, childNode, doc)
    Add namespace to the element node in libxml structure from opensoap structure 
    
    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm_child|))
        OpenSOAP Element Pointer
      :xmlNodePtr [in, out] ((|childNode|))
        Libxml Node Pointer
      :xmlDocPtr [in ] ((|doc|))
        Libxml Document Pointer
    :Return value
      :type int
        error code.
      
=end
*/
static
int
OpenSOAPXMLElmSetNsToLibxmlElm(/* [in]  */ OpenSOAPXMLElmPtr elm,
							   /* [out] */ xmlNodePtr childNode,
							   /* [out] */ xmlDocPtr  doc) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;
	if (elm && childNode && doc) {
		xmlNsPtr childNs = NULL;
		char *utf8NsURI = NULL;
		char *utf8NsPrefix = NULL;
		OpenSOAPXMLNamespacePtr elmsNamespace
			= ((OpenSOAPXMLNodePtr)elm)->thisNamespace;

		if (elmsNamespace) {
			int isParentDefinedNs = 0;
			ret = OpenSOAPXMLElmIsParentDefinedNs(elm,
												  elmsNamespace,
												  &isParentDefinedNs);
			if (OPENSOAP_SUCCEEDED(ret)) {
				ret = OpenSOAPXMLNamespaceGetUTF8Properties(elmsNamespace,
															NULL,
															&utf8NsURI,
															&utf8NsPrefix);
				if (OPENSOAP_SUCCEEDED(ret)) {
					xmlNsPtr	childNs = NULL;
					/* check weather there is a namespace defined */
					xmlNsPtr searchNs = xmlSearchNsByHref(doc,
														  childNode,
														  utf8NsURI);
					/* NS is parent Defined, NsURI set to NULL */
					if (isParentDefinedNs) {
						free(utf8NsURI);
						utf8NsURI = NULL;
					}

					if (searchNs
						&& (!searchNs->prefix || !utf8NsPrefix
							 || ((strcmp(searchNs->prefix, utf8NsPrefix) == 0)
								  && !isParentDefinedNs))) {
						childNs = searchNs;
					}
				}
			}
		}
		else {
			/* add NULL namespace to this element */
			ret = OPENSOAP_NO_ERROR;
		}

		if (OPENSOAP_SUCCEEDED(ret)) {
			if (!childNs) {
				childNs = xmlNewNs(childNode, utf8NsURI, utf8NsPrefix);
			}

			xmlSetNs(childNode, childNs);
		}
		/* free utf8 string */
		free(utf8NsURI);
		free(utf8NsPrefix);
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddNsDefsToLibxmlElm(childElm, childNode, doc)
    Add namespace definition to the element node in libxml structure from opensoap structure 
    
    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm_child|))
        OpenSOAP Element Pointer
      :xmlNodePtr [in, out] ((|childNode|))
        Libxml Node Pointer
      :xmlDocPtr [in ] ((|doc|))
        Libxml Document Pointer
    :Return value
      :type int
        error code.
      
=end
*/
static
int
OpenSOAPXMLElmAddNsDefsToLibxmlElm(/* [in]  */ OpenSOAPXMLElmPtr elm,
								   /* [out] */ xmlNodePtr childNode,
								   /* [out] */ xmlDocPtr  doc) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;
	if (elm && childNode && doc) {
		char *utf8NsURI = NULL;
		char *utf8NsPrefix = NULL;
		OpenSOAPXMLNamespacePtr nsItr = elm->definedXMLNsList;

		for (ret = OPENSOAP_NO_ERROR; nsItr; nsItr = nsItr->next) {
			ret = OpenSOAPXMLNamespaceGetUTF8Properties(nsItr,
														NULL,
														&utf8NsURI,
														&utf8NsPrefix);
			if (OPENSOAP_FAILED(ret)) {
				break;
			}

			if (utf8NsURI) {
				/* check weather there is a namespace defined */
				xmlNsPtr searchNs = xmlSearchNsByHref(doc,
													  childNode,
													  utf8NsURI);
				if (!searchNs
					|| (searchNs->prefix && utf8NsPrefix
						&& (strcmp(searchNs->prefix, utf8NsPrefix)
							!= 0))) {
					xmlNsPtr childNs = xmlNewNs(childNode,
												utf8NsURI,
												utf8NsPrefix);
				}
			}
			/* */
			free(utf8NsURI);
			utf8NsURI = NULL;
			free(utf8NsPrefix);
			utf8NsPrefix = NULL;
		}
		/* */
		free(utf8NsURI);
		free(utf8NsPrefix);
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddNsInEncodingString(doc, childElm, childNode)
    Add namespace to the element node in libxml structure from opensoap structure 
    
    :Parameters
      :xmlDocPtr [in ] ((|doc|))
        Libxml Document Pointer
      :OpenSOAPXMLElmPtr [in] ((|childElm|))
        OpenSOAP Element Pointer
      :xmlNodePtr [in, out] ((|childNode|))
        Libxml Node Pointer
    :Return value
      :type int
        error code.
      
=end
*/
int
OpenSOAPXMLElmAddNsInEncodingString(xmlDocPtr doc,
									OpenSOAPXMLElmPtr childElm,
									xmlNodePtr childNode) {
	/* */
	int ret = OpenSOAPXMLElmSetNsToLibxmlElm(childElm, childNode, doc);

	if (OPENSOAP_SUCCEEDED(ret)) {
		ret = OpenSOAPXMLElmAddNsDefsToLibxmlElm(childElm, childNode, doc);
	}
	
	return ret;
}	

/*
=begin
--- function#OpenSOAPXMLElmCreateChildNodeFromLibxml(elm, node)
    create opensoap structure from libxml structure 
    
    :Parameters
      :OpenSOAPXMLElmPtr [in, out] ((|elm|))
        OpenSOAP Element Pointer
      :xmlNodePtr [in] ((|node|))
        Libxml Node Pointer
    :Return value
      :type int
        error code.
      
=end
*/
int
OpenSOAPXMLElmCreateChildNodeFromLibxml(/* [in, out] */ OpenSOAPXMLElmPtr elm,
										/* [in]      */ xmlNodePtr node) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (elm && node) {
		xmlNodePtr	childNode = node->children;
		for (ret = OPENSOAP_NO_ERROR; childNode; childNode = childNode->next) {
			const char XMLNODE_TYPE_TEXT[] = "text";
			if (xmlIsBlankNode(childNode)) {
				continue;
			}
			if (childNode->name
				&& strcmp(childNode->name, XMLNODE_TYPE_TEXT) == 0) {
				/* text node */
				OpenSOAPStringPtr contentStr = NULL;
				ret = OpenSOAPStringCreateWithUTF8(childNode->content,
												   &contentStr);
				if (OPENSOAP_SUCCEEDED(ret)) {
					/* */
					OpenSOAPStringRelease(((OpenSOAPXMLNodePtr)elm)->value);
					((OpenSOAPXMLNodePtr)elm)->value = contentStr;
				}
			}
			else {
				/* not text node */
				OpenSOAPStringPtr childName = NULL;
				OpenSOAPXMLElmPtr childElm = NULL;
				/* create node name */
				ret = OpenSOAPStringCreateWithUTF8(childNode->name,
												   &childName);
				if (OPENSOAP_FAILED(ret)) {
					break;
				}
				
				/* add child element */
				ret = OpenSOAPXMLElmAddChildString(elm,
												   childName,
												   &childElm);
				OpenSOAPStringRelease(childName);
				if (OPENSOAP_FAILED(ret)) {
					break;
				}
				
				/* add element namespace */
				ret = OpenSOAPXMLElmAddNsInCharEncoding(childNode, childElm);
				if (OPENSOAP_FAILED(ret)) {
					break;
				}

				/* add element's attributes */
				ret = OpenSOAPXMLElmAddAttrInCharEncoding(childNode,
														  childElm);
				if (OPENSOAP_FAILED(ret)) {
					break;
				}
			
				/* add element's child */
				ret = OpenSOAPXMLElmCreateChildNodeFromLibxml(childElm,
															  childNode);
				if (OPENSOAP_FAILED(ret)) {
					break;
				}
			}
		}
	}
	
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( name, mb_name, mb_name_beg )
    Convert encoding from utf8-enc libxml to local encoding
    
    :Parameters
      :const xmlChar * [in ] ((|name|))
        Libxml Charactor Pointer
      :OpenSOAPByteArrayPtr [in] ((|mb_name|))
        OpenSOAP Byte Array Pointer
      :const unsigned char ** [in] ((|mb_name_beg|))
        Const Unsigned Charactor Pointer's Pointer
    :Return value
      :type int
        error code.
      
=end
*/
int
OpenSOAPXMLElmConvertEncodingFromUTF8Libxml(const xmlChar *name,
											OpenSOAPByteArrayPtr mb_name,
											const unsigned char **
											mb_name_beg) {
	int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPByteArrayPtr  utf8_name = NULL;

	unsigned char null_char = '\0';
	size_t mb_name_size = 0;

	if (name == NULL && mb_name == NULL) {
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	ret = OpenSOAPByteArrayCreateWithData(name,
										  name ? strlen(name) : 0,
										  &utf8_name);
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}

	ret = OpenSOAPStringConvertCharEncoding("UTF-8",
											utf8_name,
											NULL,
											mb_name);
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}


	ret = OpenSOAPByteArrayAppend(mb_name, &null_char, sizeof(null_char));
	if (OPENSOAP_FAILED(ret)) {
		return ret;
	}


	ret = OpenSOAPByteArrayGetBeginSizeConst( mb_name,
											  mb_name_beg,
											  &mb_name_size );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	ret = OpenSOAPByteArrayRelease( utf8_name );
	if( OPENSOAP_FAILED(ret) ){
		return ret;
	}

	ret = OPENSOAP_NO_ERROR;
	
	return ret;
}


/*
=begin
--- function#OpenSOAPXMLElmAddAttrInCharEncoding( childNode, elm_cld )
    Add attribute to the element in opensoap structure from libxml structure  
    
    :Parameters
      :xmlNodePtr [in, out] ((|childNode|))
        Libxml Node Pointer
      :OpenSOAPXMLElmPtr [in] ((|elm_cld|))
        OpenSOAP Element Pointer
    :Return value
      :type int
        error code.
      
=end
*/
int
OpenSOAPXMLElmAddAttrInCharEncoding( xmlNodePtr childNode, OpenSOAPXMLElmPtr elm_cld ){
	int ret = OPENSOAP_YET_IMPLEMENTATION;

	xmlAttrPtr nextAttr = NULL;
	xmlNsPtr   nextNs = NULL;

	OpenSOAPStringPtr attr_value = NULL;
	OpenSOAPXMLAttrPtr new_attr = NULL;


	/* add element's attributes */
	if( childNode->properties != NULL ){
		nextAttr = childNode->properties;
		while( nextAttr != NULL ){
			OpenSOAPByteArrayPtr mb_name = NULL;
			const unsigned char *mb_name_beg = NULL;
			OpenSOAPByteArrayPtr mb_content = NULL;
			const unsigned char *mb_content_beg = NULL;

			if( nextAttr->name != NULL ){
				ret = OpenSOAPByteArrayCreate( &mb_name );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextAttr->name,
																	 mb_name,
																	 &mb_name_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				mb_name = NULL;
				mb_name_beg = NULL;
			}

			if( nextAttr->children->content != NULL ){
				ret = OpenSOAPByteArrayCreate( &mb_content );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextAttr->children->content,
																	 mb_content,
																	 &mb_content_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				mb_content = NULL;
				mb_content_beg = NULL;
			}

			if( mb_content_beg != NULL ){
				ret = OpenSOAPStringCreateWithMB( mb_content_beg, &attr_value );
				if( OPENSOAP_FAILED(ret) ) {
					return ret;
				}
			}
			else{
				attr_value = NULL;
			}

			new_attr = NULL;
			ret = OpenSOAPXMLElmAddAttributeMB( elm_cld, 
												mb_name_beg,
												"string", 
												&attr_value,
												&new_attr );
			if( OPENSOAP_FAILED(ret) ) {
				return ret;
			}

			if( mb_name != NULL ){ 
				ret = OpenSOAPByteArrayRelease( mb_name );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				mb_name = NULL;
			}

			if( mb_content != NULL ){
				ret = OpenSOAPByteArrayRelease( mb_content );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				mb_content = NULL;
			}

			/* add attributes' namespace */
			if( nextAttr->ns != NULL ){
				OpenSOAPByteArrayPtr mb_href = NULL;
				const unsigned char *mb_href_beg = NULL;
				OpenSOAPByteArrayPtr mb_prefix = NULL;
				const unsigned char *mb_prefix_beg = NULL;

				nextNs = nextAttr->ns;

				if( nextNs->href != NULL ){
					ret = OpenSOAPByteArrayCreate( &mb_href );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextNs->href,
																		 mb_href,
																		 &mb_href_beg );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
				}
				else{
					mb_href = NULL;
					mb_href_beg = NULL;
				}

				if( nextNs->prefix != NULL ){
					ret = OpenSOAPByteArrayCreate( &mb_prefix );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextNs->prefix,
																		 mb_prefix,
																		 &mb_prefix_beg );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
				}
				else{
					mb_prefix = NULL;
					mb_prefix_beg = NULL;
				}

				ret = OpenSOAPXMLAttrSetNamespaceMB( new_attr,
													 mb_href_beg,
													 mb_prefix_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}

				if( mb_href != NULL ){
					ret = OpenSOAPByteArrayRelease( mb_href );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					mb_href = NULL;
				}
				if( mb_prefix != NULL ){
					ret = OpenSOAPByteArrayRelease( mb_prefix );
					if( OPENSOAP_FAILED(ret) ){
						return ret;
					}
					mb_prefix = NULL;
				}
			}
			nextAttr = nextAttr->next;
		}
	}
	else{
		ret = OPENSOAP_NO_ERROR;
	}
	return ret;
}


/*
=begin
--- function#OpenSOAPXMLElmAddNsInCharEncoding( childNode, elm_cld )
    Add namespace to the element in opensoap structure from libxml structure  
    
    :Parameters
      :xmlNodePtr [in, out] ((|childNode|))
        Libxml Node Pointer
      :OpenSOAPXMLElmPtr [in] ((|elm_cld|))
        OpenSOAP Element Pointer
    :Return value
      :type int
        error code.
      
=end
*/
int
OpenSOAPXMLElmAddNsInCharEncoding( xmlNodePtr childNode, OpenSOAPXMLElmPtr elm_cld ){
	int ret = OPENSOAP_YET_IMPLEMENTATION;

	xmlNsPtr   nextNs = NULL;

	OpenSOAPXMLNamespacePtr ns_def = NULL;

	/* add the associated namespace of this element */
	if( childNode->ns != NULL ){
		OpenSOAPByteArrayPtr mb_href = NULL;
		const unsigned char *mb_href_beg = NULL;
		OpenSOAPByteArrayPtr mb_prefix = NULL;
		const unsigned char *mb_prefix_beg = NULL;

		nextNs = childNode->ns;

		if( nextNs->href != NULL ){
			ret = OpenSOAPByteArrayCreate( &mb_href );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextNs->href,
																 mb_href,
																 &mb_href_beg );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
		}
		else{
			mb_href = NULL;
			mb_href_beg = NULL;
		}

		if( nextNs->prefix != NULL ){
			ret = OpenSOAPByteArrayCreate( &mb_prefix );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextNs->prefix,
																 mb_prefix,
																 &mb_prefix_beg );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
		}
		else{
			mb_prefix = NULL;
			mb_prefix_beg = NULL;
		}

		ret = OpenSOAPXMLElmSetNamespaceMB( elm_cld,
											mb_href_beg,
											mb_prefix_beg );
		if( OPENSOAP_FAILED(ret) ){
			return ret;
		}

		if( mb_href ){
			ret = OpenSOAPByteArrayRelease( mb_href );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			mb_href = NULL;
		}
		if( mb_prefix ){
			ret = OpenSOAPByteArrayRelease( mb_prefix );
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}
			mb_prefix = NULL;
		}
	}

	/* add namespace definitions on this element */
	if( childNode->nsDef != NULL ){
		nextNs = childNode->nsDef;
		while( nextNs != NULL ){
			OpenSOAPByteArrayPtr mb_href = NULL;
			const unsigned char *mb_href_beg = NULL;
			OpenSOAPByteArrayPtr mb_prefix = NULL;
			const unsigned char *mb_prefix_beg = NULL;

			if( nextNs->href != NULL ){
				ret = OpenSOAPByteArrayCreate( &mb_href );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextNs->href,
																	 mb_href,
																	 &mb_href_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				mb_href = NULL;
				mb_href_beg = NULL;
			}

			if( nextNs->prefix != NULL ){
				ret = OpenSOAPByteArrayCreate( &mb_prefix );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				ret = OpenSOAPXMLElmConvertEncodingFromUTF8Libxml( nextNs->prefix,
																	 mb_prefix,
																	 &mb_prefix_beg );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
			}
			else{
				mb_prefix = NULL;
				mb_prefix_beg = NULL;
			}

			ret = OpenSOAPXMLElmDefineNamespaceMB( elm_cld,
												   mb_href_beg,
												   mb_prefix_beg,
												   &ns_def);
			if( OPENSOAP_FAILED(ret) ){
				return ret;
			}

			if( mb_href ){
				ret = OpenSOAPByteArrayRelease( mb_href );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				mb_href = NULL;
			}
			if( mb_prefix ){
				ret = OpenSOAPByteArrayRelease( mb_prefix );
				if( OPENSOAP_FAILED(ret) ){
					return ret;
				}
				mb_prefix = NULL;
			}
			nextNs = nextNs->next;
		}
	}
	else{
		ret = OPENSOAP_NO_ERROR;
	}
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetNameString(elm, name)
    Getting of XML Element's Name as OpenSOAPString

    :Parameters
      :[in]  OpenSOAPXMLElmPtr ((|elm|))
        XML Element Pointer.
      :[out] OpenSOAPStringPtr * ((|name|))
        Result Name of Node.
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
int
OPENSOAP_API
OpenSOAPXMLElmGetNameString(/* [in]  */ OpenSOAPXMLElmPtr elm,
							/* [out] */ OpenSOAPStringPtr *name) {
	int ret = OpenSOAPXMLNodeGetName((OpenSOAPXMLNodePtr)elm, name);

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetNextAttr(elm, attr_elm)
    Get next attr XML Element.

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :OpenSOAPXMLAttrPtr * [in, out] ((|attr_elm|))
        Next Attr Element Pointer of Assignment XML Element
        But If *attr_elm as in NULL then Return To First XML Element Pointer 
    :Return Value value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetNextAttr(OpenSOAPXMLElmPtr /* [in] */ elm,
                           OpenSOAPXMLAttrPtr * /* [in, out] */ attr_elm) {
    int ret = OPENSOAP_YET_IMPLEMENTATION;

	OpenSOAPXMLNodePtr  attr_next;

	if( !elm ){
		ret = OPENSOAP_PARAMETER_BADVALUE;
		return ret;
	}

	if( !(*attr_elm) ){
		*attr_elm = (OpenSOAPXMLAttrPtr)(elm->xmlAttrList);
		ret = OPENSOAP_NO_ERROR;
	}
	else{
		attr_next = (OpenSOAPXMLNodePtr)(elm->xmlAttrList);
		while( attr_next != NULL ){
			if( attr_next == (OpenSOAPXMLNodePtr)(*attr_elm) ){
				*attr_elm = (OpenSOAPXMLAttrPtr)((*attr_elm)->super.next);
				ret = OPENSOAP_NO_ERROR;
				return ret;
			}
			attr_next = attr_next->next;
		}
	}
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetChildValueString(elm, pName, typeName, value)
    Get Child Value

    :Parameters
      :[in]  OpenSOAPXMLElmPtr ((|elm|))
        OpenSOAP XML Element.
      :[in]  const char * ((|childName|))
        child element name
      :[in]  const char * ((|typeName|))
        value's type name
      :[out] void * ((|value|))
        value buffer.
    :Return Value value
      :int
	    error code.
=end
 */
static
int
OpenSOAPXMLElmGetChildValueString(/* [in]  */ OpenSOAPXMLElmPtr elm,
								  /* [in]  */ OpenSOAPStringPtr childName,
								  /* [in]  */ OpenSOAPStringPtr typeName,
								  /* [out] */ void *value) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (elm && childName && typeName && value) {
		OpenSOAPXMLElmPtr childElm = NULL;
		ret = OpenSOAPXMLElmGetChildString(elm, childName, &childElm);
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = OpenSOAPXMLElmGetValueString(childElm, typeName, value);
		}
	}

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetChildValueString(elm, pName, typeName, value)
    Set Child Value

    :Parameters
      :[out] OpenSOAPXMLElmPtr ((|elm|))
        OpenSOAP XML Element.
      :[in]  const wchar_t * ((|childName|))
        child element name
      :[in]  const wchar_t * ((|typeName|))
        value's type name
      :[in]  void * ((|value|))
        value buffer.
    :Return Value value
      :int
	    error code.
=end
 */
static
int
OpenSOAPXMLElmSetChildValueString(/* [out] */ OpenSOAPXMLElmPtr elm,
								  /* [in]  */ OpenSOAPStringPtr childName,
								  /* [in]  */ OpenSOAPStringPtr typeName,
								  /* [in]  */ void *value) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (elm && childName && typeName && value) {
		OpenSOAPXMLElmPtr childElm = NULL;
		ret = OpenSOAPXMLElmGetChildString(elm, childName, &childElm);
		if (OPENSOAP_SUCCEEDED(ret)) {
			if (!childElm) {
				ret = OpenSOAPXMLElmAddChildString(elm, childName, &childElm);
			}
			if (OPENSOAP_SUCCEEDED(ret)) {
				ret = OpenSOAPXMLElmSetValueString(childElm, typeName, value);
			}
		}
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetChildValueMB(elm, pName, typeName, value)
    Get Child Value

    :Parameters
      :[in]  OpenSOAPXMLElmPtr ((|elm|))
        OpenSOAP XML Element.
      :[in]  const char * ((|childName|))
        child element name
      :[in]  const char * ((|typeName|))
        value's type name
      :[out] void * ((|value|))
        value buffer.
    :Return Value value
      :int
	    error code.
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetChildValueMB(/* [in]  */ OpenSOAPXMLElmPtr elm,
							  /* [in]  */ const char *childName,
							  /* [in]  */ const char *typeName,
							  /* [out] */ void *value) {
	OpenSOAPStringPtr childNameStr = NULL;
	/* create child name string */
	int ret = OpenSOAPStringCreateWithMB(childName, &childNameStr);
	if (OPENSOAP_SUCCEEDED(ret)) {
		OpenSOAPStringPtr typeNameStr = NULL;
		/* create type name string */
		ret = OpenSOAPStringCreateWithMB(typeName, &typeNameStr);
		if (OPENSOAP_SUCCEEDED(ret)) {
			/* get value */
			ret = OpenSOAPXMLElmGetChildValueString(elm,
													childNameStr,
													typeNameStr,
													value);
			/* release type name string */
			OpenSOAPStringRelease(typeNameStr);
		}
		/* release child name string */
		OpenSOAPStringRelease(childNameStr);
	}
	
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmGetChildValueWC(elm, pName, typeName, value)
    Get Child Value

    :Parameters
      :[in]  OpenSOAPXMLElmPtr ((|elm|))
        OpenSOAP XML Element.
      :[in]  const wchar_t * ((|childName|))
        child element name
      :[in]  const wchar_t * ((|typeName|))
        value's type name
      :[out] void * ((|value|))
        value buffer.
    :Return Value value
      :int
	    error code.
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmGetChildValueWC(/* [in]  */ OpenSOAPXMLElmPtr elm,
							  /* [in]  */ const wchar_t *childName,
							  /* [in]  */ const wchar_t *typeName,
							  /* [out] */ void *value) {
	OpenSOAPStringPtr childNameStr = NULL;
	/* create child name string */
	int ret = OpenSOAPStringCreateWithWC(childName, &childNameStr);
	if (OPENSOAP_SUCCEEDED(ret)) {
		OpenSOAPStringPtr typeNameStr = NULL;
		/* create type name string */
		ret = OpenSOAPStringCreateWithWC(typeName, &typeNameStr);
		if (OPENSOAP_SUCCEEDED(ret)) {
			/* get value */
			ret = OpenSOAPXMLElmGetChildValueString(elm,
													childNameStr,
													typeNameStr,
													value);
			/* release type name string */
			OpenSOAPStringRelease(typeNameStr);
		}
		/* release child name string */
		OpenSOAPStringRelease(childNameStr);
	}
	
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetChildValueMB(elm, pName, typeName, value)
    Set Child Value

    :Parameters
      :[out] OpenSOAPXMLElmPtr ((|elm|))
        OpenSOAP XML Element.
      :[in]  const char * ((|childName|))
        child element name
      :[in]  const char * ((|typeName|))
        value's type name
      :[in]  void * ((|value|))
        value buffer.
    :Return Value value
      :int
	    error code.
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetChildValueMB(/* [out] */ OpenSOAPXMLElmPtr elm,
							  /* [in]  */ const char *childName,
							  /* [in]  */ const char *typeName,
							  /* [in]  */ void *value) {
	OpenSOAPStringPtr childNameStr = NULL;
	/* create child name string */
	int ret = OpenSOAPStringCreateWithMB(childName, &childNameStr);
	if (OPENSOAP_SUCCEEDED(ret)) {
		OpenSOAPStringPtr typeNameStr = NULL;
		/* create type name string */
		ret = OpenSOAPStringCreateWithMB(typeName, &typeNameStr);
		if (OPENSOAP_SUCCEEDED(ret)) {
			/* set value */
			ret = OpenSOAPXMLElmSetChildValueString(elm,
													childNameStr,
													typeNameStr,
													value);
			/* release type name string */
			OpenSOAPStringRelease(typeNameStr);
		}
		/* release child name string */
		OpenSOAPStringRelease(childNameStr);
	}
	
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetChildValueWC(elm, pName, typeName, value)
    Set Child Value

    :Parameters
      :[out] OpenSOAPXMLElmPtr ((|elm|))
        OpenSOAP XML Element.
      :[in]  const wchar_t * ((|childName|))
        child element name
      :[in]  const wchar_t * ((|typeName|))
        value's type name
      :[in]  void * ((|value|))
        value buffer.
    :Return Value value
      :int
	    error code.
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetChildValueWC(/* [out] */ OpenSOAPXMLElmPtr elm,
							  /* [in]  */ const wchar_t *childName,
							  /* [in]  */ const wchar_t *typeName,
							  /* [in]  */ void *value) {
	OpenSOAPStringPtr childNameStr = NULL;
	/* create child name string */
	int ret = OpenSOAPStringCreateWithWC(childName, &childNameStr);
	if (OPENSOAP_SUCCEEDED(ret)) {
		OpenSOAPStringPtr typeNameStr = NULL;
		/* create type name string */
		ret = OpenSOAPStringCreateWithWC(typeName, &typeNameStr);
		if (OPENSOAP_SUCCEEDED(ret)) {
			/* set value */
			ret = OpenSOAPXMLElmSetChildValueString(elm,
													childNameStr,
													typeNameStr,
													value);
			/* release type name string */
			OpenSOAPStringRelease(typeNameStr);
		}
		/* release child name string */
		OpenSOAPStringRelease(childNameStr);
	}
	
	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetValueAsString(elm, name, isDup)
    XML Element value set as string

    :Parameters
      :[out] OpenSOAPXMLElmPtr ((|elm|))
        XML Elm Pointer.
      :[in]  OpenSOAPStringPtr ((|value|))
        Setting Elm Name
      :[in]  int ((|isDup|))
        duplicate flag. If non zero, then value is duplicate.
    :Return Value
      :int
	    error code.
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetValueAsString(/* [out] */ OpenSOAPXMLElmPtr elm,
							   /* [in]  */ OpenSOAPStringPtr value,
							   /* [in]  */ int isDup) {
    int ret = OpenSOAPXMLNodeSetValueAsString((OpenSOAPXMLNodePtr)elm,
											  value,
											  isDup);

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetValueAsStringMB(elm, name)
    XML Element value set as string

    :Parameters
      :[out] OpenSOAPXMLElmPtr ((|elm|))
        XML Elm Pointer.
      :[in]  OpenSOAPStringPtr ((|value|))
        value as string
    :Return Value
      :int
	    error code.
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetValueAsStringMB(/* [out] */ OpenSOAPXMLElmPtr elm,
								 /* [in]  */ const char *value) {
    int ret = OpenSOAPXMLNodeSetValueAsStringMB((OpenSOAPXMLNodePtr)elm,
												value);

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetValueAsStringWC(elm, name)
    XML Element value set as string

    :Parameters
      :[out] OpenSOAPXMLElmPtr ((|elm|))
        XML Elm Pointer.
      :[in]  const wchar_t * ((|value|))
        value as string
    :Return Value
      :int
	    error code.
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetValueAsStringWC(/* [out] */ OpenSOAPXMLElmPtr elm,
								 /* [in]  */ const wchar_t *value) {
    int ret = OpenSOAPXMLNodeSetValueAsStringWC((OpenSOAPXMLNodePtr)elm,
												value);

	return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetChildValueAsString(elm, childName, value, isValueDup)
    Set Child Value

    :Parameters
      :[out] OpenSOAPXMLElmPtr ((|elm|))
        OpenSOAP XML Element.
      :[in]  OpenSOAPStringPtr ((|childName|))
        child element name
      :[in]  int ((|isValueDup|))
        value duplicate flag.
      :[in, out]  OpenSOAPStringPtr ((|value|))
        value buffer.
    :Return Value value
      :int
	    error code.
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetChildValueAsString(/* [out] */ OpenSOAPXMLElmPtr elm,
									/* [in]  */ OpenSOAPStringPtr childName,
									/* [in]  */ int isValueDup,
									/* [in, out] */ OpenSOAPStringPtr value) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (elm && childName && value) {
		OpenSOAPXMLElmPtr childElm = NULL;
		ret = OpenSOAPXMLElmGetChildString(elm, childName, &childElm);
		if (OPENSOAP_SUCCEEDED(ret)) {
			if (!childElm) {
				ret = OpenSOAPXMLElmAddChildString(elm, childName, &childElm);
			}
			if (OPENSOAP_SUCCEEDED(ret)) {
				ret = OpenSOAPXMLElmSetValueAsString(childElm,
													 value,
													 isValueDup);
			}
		}
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetChildValueAsStringMB(elm, childName, value)
    Set Child Value

    :Parameters
      :[out] OpenSOAPXMLElmPtr ((|elm|))
        OpenSOAP XML Element.
      :[in]  const char * ((|childName|))
        child element name
      :[in]  const char * ((|value|))
        value buffer.
    :Return Value value
      :int
	    error code.
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetChildValueAsStringMB(/* [out] */ OpenSOAPXMLElmPtr elm,
									  /* [in]  */ const char *childName,
									  /* [in]  */ const char *value) {
	OpenSOAPStringPtr childNameStr = NULL;
	int ret = OpenSOAPStringCreateWithMB(childName, &childNameStr);

	if (OPENSOAP_SUCCEEDED(ret)) {
		OpenSOAPStringPtr valueStr = NULL;
		ret = OpenSOAPStringCreateWithMB(value, &valueStr);
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = OpenSOAPXMLElmSetChildValueAsString(elm,
													  childNameStr,
													  0,
													  valueStr);
			
			if (OPENSOAP_FAILED(ret)) {
				OpenSOAPStringRelease(valueStr);
			}
		}

		OpenSOAPStringRelease(childNameStr);
	}
	
    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmSetChildValueAsStringWC(elm, childName, value)
    Set Child Value

    :Parameters
      :[out] OpenSOAPXMLElmPtr ((|elm|))
        OpenSOAP XML Element.
      :[in]  const wchar_t * ((|childName|))
        child element name
      :[in]  const wchar_t * ((|value|))
        value buffer.
    :Return Value value
      :int
	    error code.
=end
 */
extern
int
OPENSOAP_API
OpenSOAPXMLElmSetChildValueAsStringWC(/* [out] */ OpenSOAPXMLElmPtr elm,
									  /* [in]  */ const wchar_t *childName,
									  /* [in]  */ const wchar_t *value) {
	OpenSOAPStringPtr childNameStr = NULL;
	int ret = OpenSOAPStringCreateWithWC(childName, &childNameStr);

	if (OPENSOAP_SUCCEEDED(ret)) {
		OpenSOAPStringPtr valueStr = NULL;
		ret = OpenSOAPStringCreateWithWC(value, &valueStr);
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = OpenSOAPXMLElmSetChildValueAsString(elm,
													  childNameStr,
													  0,
													  valueStr);
			
			if (OPENSOAP_FAILED(ret)) {
				OpenSOAPStringRelease(valueStr);
			}
		}

		OpenSOAPStringRelease(childNameStr);
	}

    return ret;
}

/*
=begin
--- function#OpenSOAPXMLElmAddChildString(elm, cld_name, cld_elm)
    Add Child XML Element(OpenSOAPString)

    :Parameters
      :OpenSOAPXMLElmPtr [in] ((|elm|))
        OpenSOAP XML Element.
      :OpenSOAPStringPtr [in] ((|cld_name|))
        Add Child Element Name
      :OpenSOAPXMLElmPtr * [out] ((|cld_elm|))
        Add Child Element Pointer
    :Return Value value
      :int
	    Return Code Is ErrorCode
=end
*/
extern
int
OPENSOAP_API
OpenSOAPXMLElmRemoveChildString(/* [in]  */ OpenSOAPXMLElmPtr elm,
								/* [in]  */ OpenSOAPStringPtr childName,
								/* [in]  */ int isValueRelease,
								/* [out] */ OpenSOAPStringPtr *childValue) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;
	
	if (elm && childName) {
		OpenSOAPXMLElmPtr childElm = NULL;
		ret = OpenSOAPXMLElmGetChildString(elm, childName, &childElm);
		if (OPENSOAP_SUCCEEDED(ret)) {
			OpenSOAPXMLNodePtr childNode = (OpenSOAPXMLNodePtr)childElm;
			ret = OpenSOAPXMLNodeRemoveValue(childNode,
											 isValueRelease,
											 childValue);
			if (OPENSOAP_SUCCEEDED(ret)) {
				ret = OpenSOAPXMLNodeRemove(childNode);
			}
		}
	}
	
    return ret;
}

